# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from pyparsing import Literal, Word, nums, alphas, OneOrMore, ZeroOrMore, restOfLine, LineEnd, Combine, White
from pysmt.shortcuts import TRUE, FALSE, And, Or, Symbol, BV, EqualsOrIff, Implies, get_env
from pysmt.typing import BOOL, BVType
from pysmt.parsing import parse, HRParser, HRLexer, PrattParser, Rule, UnaryOpAdapter, InfixOpAdapter

import pysmt.environment
import pysmt.formula

from cosa.encoders.formulae import StringParser
from cosa.utils.logger import Logger

from pysmt.operators import new_node_type

LTL_X = new_node_type(node_str="LTL_X")
LTL_F = new_node_type(node_str="LTL_F")
LTL_G = new_node_type(node_str="LTL_G")
LTL_U = new_node_type(node_str="LTL_U")
LTL_R = new_node_type(node_str="LTL_R")
LTL_O = new_node_type(node_str="LTL_O")
LTL_H = new_node_type(node_str="LTL_H")
ALL_LTL = (LTL_X, LTL_F, LTL_G, LTL_U, LTL_R, LTL_O, LTL_H)

ASSIGN = new_node_type(node_str="ASSIGN")
DEFINE = new_node_type(node_str="DEFINE")

def Assign(left, right):
    return get_env().formula_manager.Assign(left, right)

def Define(left, right):
    return get_env().formula_manager.Define(left, right)

def Finally(arg):
    return get_env().formula_manager.F(arg)

class FormulaManager(pysmt.formula.FormulaManager):
    """Extension of FormulaManager to handle LTL Operators."""
    def X(self, formula):
        return self.create_node(node_type=LTL_X, args=(formula,))
    
    def F(self, formula):
        return self.create_node(node_type=LTL_F, args=(formula,))

    def G(self, formula):
        return self.create_node(node_type=LTL_G, args=(formula,))

    def U(self, left, right):
        return self.create_node(node_type=LTL_U, args=(left, right))

    def R(self, left, right):
        return self.create_node(node_type=LTL_R, args=(left, right))

    def O(self, formula):
        return self.create_node(node_type=LTL_O, args=(formula,))

    def H(self, formula):
        return self.create_node(node_type=LTL_H, args=(formula,))

    def Assign(self, left, right):
        return self.create_node(node_type=ASSIGN, args=(left, right), payload=(1,))

    def Define(self, left, right):
        return self.create_node(node_type=DEFINE, args=(left, right), payload=(1,))
    
class ExtLexer(HRLexer):
    def __init__(self, env=None):
        HRLexer.__init__(self, env=env)
        self.rules.insert(0, Rule(r"(!=)", InfixOpAdapter(self.NEquals, 60), False))
        self.rules.insert(0, Rule(r"(X)", UnaryOpAdapter(self.X, 100), False))
        self.rules.insert(0, Rule(r"(F)", UnaryOpAdapter(self.F, 100), False))
        self.rules.insert(0, Rule(r"(G)", UnaryOpAdapter(self.G, 100), False))
        self.rules.insert(0, Rule(r"(U)", InfixOpAdapter(self.U, 100), False))
        self.rules.insert(0, Rule(r"(R)", InfixOpAdapter(self.R, 100), False))
        self.rules.insert(0, Rule(r"(O)", UnaryOpAdapter(self.O, 100), False))
        self.rules.insert(0, Rule(r"(H)", UnaryOpAdapter(self.H, 100), False))
        self.compile()

    def X(self, x):
        return self.mgr.X(x)
    
    def F(self, x):
        return self.mgr.F(x)
    
    def G(self, x):
        return self.mgr.G(x)

    def U(self, l, r):
        return self.mgr.U(l, r)

    def R(self, l, r):
        return self.mgr.R(l, r)

    def O(self, x):
        return self.mgr.O(x)

    def H(self, x):
        return self.mgr.H(x)
    
    def NEquals(self, l, r):
        return self.mgr.Not(self.mgr.Equals(l, r))

    
from pysmt.type_checker import SimpleTypeChecker
SimpleTypeChecker.set_handler(SimpleTypeChecker.walk_bool_to_bool, *ALL_LTL)
SimpleTypeChecker.set_handler(SimpleTypeChecker.walk_bv_to_bool, ASSIGN)
SimpleTypeChecker.set_handler(SimpleTypeChecker.walk_bv_to_bool, DEFINE)

from pysmt.oracles import FreeVarsOracle
FreeVarsOracle.set_handler(FreeVarsOracle.walk_simple_args, *ALL_LTL)
FreeVarsOracle.set_handler(FreeVarsOracle.walk_simple_args, ASSIGN)
FreeVarsOracle.set_handler(FreeVarsOracle.walk_simple_args, DEFINE)

import pysmt.printers
from pysmt.walkers.generic import handles
LTL_TYPE_TO_STR = { LTL_X: "X", LTL_F: "F", LTL_G: "G", LTL_O: "O", LTL_H: "H"}

class ExtHRPrinter(pysmt.printers.HRPrinter):
    def walk_ltl_r(self, formula):
        return self.walk_nary(formula, " R ")

    def walk_ltl_u(self, formula):
        return self.walk_nary(formula, " U ")

    def walk_assign(self, formula):
        return self.walk_nary(formula, " <= ")

    def walk_define(self, formula):
        return self.walk_nary(formula, " = ")
    
    @handles(LTL_X, LTL_F, LTL_G, LTL_O, LTL_H)
    def walk_ltl(self, formula):
        node_type = formula.node_type()
        op_symbol = LTL_TYPE_TO_STR[node_type]
        self.stream.write("(%s " % op_symbol)
        yield formula.arg(0)
        self.stream.write(")")

# EOC HRPrinter

class HRSerializer(pysmt.printers.HRSerializer):
    PrinterClass = ExtHRPrinter

# EOC HRSerialize

from pysmt.walkers import IdentityDagWalker

def walk_ltl_x(self, formula, args, **kwargs): return self.mgr.X(args[0])
def walk_ltl_f(self, formula, args, **kwargs): return self.mgr.F(args[0])
def walk_ltl_g(self, formula, args, **kwargs): return self.mgr.G(args[0])
def walk_ltl_u(self, formula, args, **kwargs): return self.mgr.U(args[0], args[1])
def walk_ltl_r(self, formula, args, **kwargs): return self.mgr.R(args[0], args[1])
def walk_ltl_o(self, formula, args, **kwargs): return self.mgr.O(args[0])
def walk_ltl_h(self, formula, args, **kwargs): return self.mgr.H(args[0])

def walk_assign(self, formula, args, **kwargs): return self.mgr.Assign(args[0], args[1])
def walk_define(self, formula, args, **kwargs): return self.mgr.Define(args[0], args[1])

IdentityDagWalker.set_handler(walk_ltl_x, LTL_X)
IdentityDagWalker.set_handler(walk_ltl_f, LTL_F)
IdentityDagWalker.set_handler(walk_ltl_g, LTL_G)
IdentityDagWalker.set_handler(walk_ltl_u, LTL_U)
IdentityDagWalker.set_handler(walk_ltl_r, LTL_R)
IdentityDagWalker.set_handler(walk_ltl_o, LTL_O)
IdentityDagWalker.set_handler(walk_ltl_h, LTL_H)

IdentityDagWalker.set_handler(walk_assign, ASSIGN)
IdentityDagWalker.set_handler(walk_define, DEFINE)

def walk_simplify_assign(self, formula, args, **kwargs):
    assert len(args) == 2

    sl = args[0]
    sr = args[1]

    if sl.is_constant() and sr.is_constant():
        l = sl.constant_value()
        r = sr.constant_value()
        return self.manager.Bool(l == r)
    elif sl == sr:
        return self.manager.TRUE()
    else:
        return self.manager.Assign(sl, sr)

def walk_simplify_define(self, formula, args, **kwargs):
    assert len(args) == 2

    sl = args[0]
    sr = args[1]

    if sl.is_constant() and sr.is_constant():
        l = sl.constant_value()
        r = sr.constant_value()
        return self.manager.Bool(l == r)
    elif sl == sr:
        return self.manager.TRUE()
    else:
        return self.manager.Define(sl, sr)
    
from pysmt.simplifier import Simplifier
Simplifier.set_handler(walk_simplify_assign, ASSIGN)
Simplifier.set_handler(walk_simplify_define, DEFINE)

# EOC IdentityDagWalker

from pysmt.environment import Environment, pop_env, get_env
from pysmt.environment import push_env as pysmt_push_env

class CosaEnvironment(Environment):
    """Extension of pySMT environment."""
    FormulaManagerClass = FormulaManager
    HRSerializerClass = HRSerializer

# EOC EnvironmentLTL

def push_env(env=None):
    """Overload push_env to default to the new Environment class."""
    if env is None:
        env = CosaEnvironment()
    return pysmt_push_env(env=env)

def reset_env():
    """Overload reset_env to use the new push_env()."""
    pop_env()
    push_env()
    return get_env()





# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import math
import shutil
import copy

VPARSER = True

try:
    from pyverilog.vparser.parser import VerilogParser, VerilogPreprocessor
    from pyverilog.vparser.ply.yacc import yacc
    from pyverilog.vparser.lexer import VerilogLexer
    from pyverilog.vparser.plyparser import ParseError
    from pyverilog.vparser.ast import *
except:
    VPARSER = False

from pysmt.shortcuts import Symbol, BV, simplify, TRUE, FALSE, get_type, get_model, is_sat, is_unsat
from pysmt.shortcuts import And, Implies, Iff, Not, BVAnd, EqualsOrIff, Ite, Or, Xor, \
    BVExtract, BVAdd, BVSub, BVUDiv, BVConcat, BVULT, BVULE, BVUGT, BVUGE, BVXor, BVOr, BVLShl, BVZero, BVMul, BVNot
from pysmt.fnode import FNode
from pysmt.typing import BOOL, BVType, ArrayType, INT
from pysmt.rewritings import conjunctive_partition
from pysmt.walkers.identitydag import IdentityDagWalker
from pysmt.walkers import TreeWalker
from pysmt.walkers.dag import DagWalker

from cosa.utils.logger import Logger
from cosa.encoders.template import ModelParser, ModelInformation
from cosa.encoders.modules import Modules
from cosa.walkers.verilog_walker import VerilogWalker
from cosa.representation import HTS, TS
from cosa.utils.generic import bin_to_dec, dec_to_bin, suppress_output, restore_output
from cosa.utils.formula_mngm import B2BV, BV2B, get_free_variables, substitute, mem_access
from cosa.environment import Assign, Define, ASSIGN, DEFINE, Finally
from cosa.printers.template import HIDDEN_VAR

KEYWORDS = ""
KEYWORDS += "module wire assign else reg always endmodule end define integer generate "
KEYWORDS += "for localparam begin input output parameter posedge negedge or and if"
KEYWORDS = KEYWORDS.split()

IVERILOG = "iverilog"

MAXINT = 64

POSEDGE = "posedge"
NEGEDGE = "negedge"
LEVEL = "level"
INPUT = "input"
OUTPUT = "output"
CLOCK = "clk"

MODPARST = "__|"
MODPAREN = "|"

# SystemVerilog Verific modules
SVA_POSEDGE = "sva_posedge"
SVA_AT = "sva_at"
SVA_ASSERT = "sva_assert"
SVA_IMMEDIATE_ASSERT = "sva_immediate_assert"

SV_MODULES = [SVA_POSEDGE, SVA_AT, SVA_ASSERT, SVA_IMMEDIATE_ASSERT]

ASSERT_ST = HIDDEN_VAR+"assert__l"
ASSERTI_ST = HIDDEN_VAR+"asserti__l"
ASSERT_EN = HIDDEN_VAR[::-1]
ASSERT_SYMBOL = ASSERT_ST+"%d"+ASSERT_EN
ASSERTI_SYMBOL = ASSERTI_ST+"%d"+ASSERT_EN

NODE_ST = HIDDEN_VAR
NODE_SYMBOL = NODE_ST+"%s"

DONTCARE_VAR = HIDDEN_VAR+"value_%s"

class VerilogHTSParser(ModelParser):
    parser = None
    extensions = ["v"]
    name = "Verilog"

    files_from_dir = False
    
    def __init__(self):
        self.walker = VerilogSTSWalker()

    def get_model_info(self):
        model_info = ModelInformation()
        model_info.abstract_clock_list = self.abstract_clock_list
        model_info.clock_list = self.clock_list
        return model_info
    
    def is_available(self):
        return VPARSER and shutil.which(IVERILOG) is not None
    
    def parse_file(self, strfile, config, flags=None):
        invar_props = []
        ltl_props = []

        if flags is None:
            Logger.error("Module name not provided")
        
        absstrfile = os.path.abspath(strfile)

        print_level = 3
        if not Logger.level(print_level):
            saved_stdout = suppress_output()

        ast = self.parse([absstrfile])

        if not Logger.level(print_level):
            restore_output(saved_stdout)
        
        self.walker.config = config
        hts = self.walker.walk(ast, flags[0])
        self.abstract_clock_list = self.walker.abstract_clock_list
        self.clock_list = self.walker.clock_list
        
        hts.flatten()

        if config.zero_init:
            ts = TS("zero-init")
            assigns = []
            for var in hts.state_vars:
                if var.symbol_type().is_bv_type():
                    assigns.append(EqualsOrIff(var, BVZero(var.bv_width())))
                if var.symbol_type().is_bool_type():
                    assigns.append(Not(var))
            ts.init = And(assigns)
            hts.add_ts(ts)

        invar_props, ltl_props = self.__extract_properties(hts, strfile)

        return (hts, invar_props, ltl_props)

    def get_extensions(self):
        return self.extensions

    def parse(self, filelist, preprocess_include=None, preprocess_define=None):
        codeparser = VerilogCodeParser(filelist,
                                       preprocess_include=preprocess_include,
                                       preprocess_define=preprocess_define)
        ast = codeparser.parse()
        return ast
    
    @staticmethod        
    def get_extensions():
        return VerilogHTSParser.extensions

    def remap_an2or(self, name):
        return name

    def remap_or2an(self, name):
        return name

    def __extract_properties(self, hts, strfile):
        assertvars = [v.symbol_name() for v in hts.vars if ASSERT_ST in v.symbol_name() or ASSERTI_ST in v.symbol_name()]

        invar_props = []
        ltl_props = []

        print_line = False

        # If the assertion line has a comment in the format // <filename>(lineno)
        # it reports the line in the comment
        def extract_lineno(line, linenum):
            orig_lineno = re.search("/\w.*\(\d+\)", line)
            if orig_lineno is None:
                return line, linenum

            strfile, linenum = re.search("/\w.*\(", line), re.search("\(\d+\)", line)
            if (strfile is None) or (linenum is None):
                return line, linenum
            
            strfile = strfile.group(0)[:-1]
            linenum = int(linenum.group(0)[1:-1])

            if print_line:
                with open(strfile, "r") as f:
                    line = f.readlines()[linenum-1]

            return line, linenum
            
        
        if len(assertvars) > 0:
            with open(strfile, "r") as f:
                lines = f.readlines()
            
            for assertion in assertvars:
                asserti = ASSERTI_ST in assertion
                prefix = ASSERTI_ST if asserti else ASSERT_ST
                linenum = int(assertion[assertion.find(prefix)+len(prefix):assertion.find(ASSERT_EN)])
                line, linenum = extract_lineno(lines[linenum-1], linenum)
                line = re.sub(' +',' ',line.strip())

                lineprint = ", \"%s\""%line if print_line else ""
                
                if asserti:
                    ltl_props.append(("ImmediateAssertion_line_%d"%linenum, "Immediate assertion at line %d%s"%(linenum, lineprint), "F(%s)"%assertion))
                else:
                    invar_props.append(("Assertion_line_%d"%linenum, "Assertion at line %d%s"%(linenum, lineprint), assertion))

        return invar_props, ltl_props

class SpecVerilogParser(VerilogParser):

    def __init__(self):
        self.lexer = VerilogLexer(error_func=self._lexer_error_func)
        self.lexer.build()

        self.tokens = self.lexer.tokens
        self.parser = yacc(module=self, method="LALR", outputdir="/tmp/", debug=False)
    
class VerilogCodeParser(object):

    def __init__(self, filelist, preprocess_output='preprocess.output',
                 preprocess_include=None,
                 preprocess_define=None):
        self.preprocess_output = preprocess_output
        self.directives = ()
        self.preprocessor = VerilogPreprocessor(filelist, preprocess_output,
                                                preprocess_include,
                                                preprocess_define)
        prev_path = sys.path
        self.parser = SpecVerilogParser()
        sys.path = prev_path

    def preprocess(self):
        self.preprocessor.preprocess()
        with open(self.preprocess_output) as f:
            text = f.read()
        os.remove(self.preprocess_output)
        return text

    def parse(self, preprocess_output='preprocess.output', debug=0):
        text = self.preprocess()
        try:
            ast = self.parser.parse(text, debug=debug)
        except ParseError as e:
            Logger.error("Parsing at line %s"%(e))
        self.directives = self.parser.get_directives()
        return ast

    def get_directives(self):
        return self.directives

class VerilogSTSWalker(VerilogWalker):
    varmap = None
    paramdic = None
    subhtsdic = None
    modulesdic = None
    hts = None
    ts = None
    config = None

    reglist = None
    outputlist = None
    memlist = None
    mod_map = None
    
    id_vars = 0

    abstract_clock_list = None
    clock_list = None

    hide_autogenerated = True
    
    def __init__(self):
        self.reset_structures()

    def reset_structures(self, modulename=""):
        self.hts = HTS(modulename)
        self.ts = TS()
        self.portslist = None
        self.reglist = []
        self.outputlist = []
        self.memlist = []
        self.abstract_clock_list = []
        self.clock_list = []
        if self.varmap is None: self.varmap = {}
        if self.paramdic is None: self.paramdic = {}
        if self.modulesdic is None: self.modulesdic = {}
        self._init_mod_map()
        
    def _fresh_symbol(self, name):
        VerilogSTSWalker.id_vars += 1
        return "%s__%d"%(name, VerilogSTSWalker.id_vars)

    def assign_define_substitute(self, formula):
        subwalker = AssignDefineSubstituteWalker()
        subwalker.reset()
        ret = subwalker.walk(formula)
        return (ret, list(subwalker.assign_list))

    def is_autogenerated(self, name):
        if re.match("n\d+", name) is not None:
            return True

        if re.match("Select_\d+", name) is not None:
            return True

        if re.match("LessThan_\d+u_\d+u", name) is not None:
            return True

        if re.match("add_\d+u_\d+u", name) is not None:
            return True

        if re.match("mult_\d+s_\d+s", name) is not None:
            return True
        
        if name == "VERIFIC_DLATCHRS":
            return True

        return False
    
    def add_init(self, constraint):
        self.ts.init = And(self.ts.init, self.assign_define_substitute(constraint)[0])

    def add_invar(self, constraint):
        self.ts.invar = And(self.ts.invar, self.assign_define_substitute(constraint)[0])

    def add_trans(self, constraint):
        self.ts.trans = And(self.ts.trans, self.assign_define_substitute(constraint)[0])
        
    def add_constraint(self, constraint):
        if TS.has_next(constraint):
            self.add_trans(constraint)
        else:
            self.add_invar(constraint)

    def add_ftrans(self, variable, cond_assign_list):
        self.ts.add_func_trans(variable, cond_assign_list)

    def clean_name(self, name):
        if name[0] == "\\":
            return name[1:]
        return name
            
    def varname(self, modulename, varname):
        varname = self.clean_name(varname)
        if modulename == "":
            return varname
        return "%s.%s"%(modulename, varname)

    def _add_clock_behavior(self, var, modulename):
        varname = var.symbol_name()
        if (CLOCK in varname):
            self.clock_list.append(var)

    def _init_mod_map(self):
        mod_map = []
        mod_map.append(("and",  Modules.And_M))
        mod_map.append(("nand",  Modules.Nand_M))
        mod_map.append(("or",   Modules.Or_M))
        mod_map.append(("nor",  Modules.Nor_M))
        mod_map.append(("not",  Modules.Not_M))
        mod_map.append(("xor",  Modules.Xor_M))

        self.mod_map = dict(mod_map)
            
    def Paramlist(self, modulename, el, args):
        return el

    def Wire(self, modulename, el, args):
        elname = el.name

        if self.hide_autogenerated:
            if self.is_autogenerated(el.name):
                elname = NODE_SYMBOL%elname

        width = args[0] if args is not None else 1
        wvar = Symbol(self.varname(modulename, elname), BVType(width))
        self.add_var(modulename, elname, wvar)
        self.ts.add_var(wvar)
        return (elname, args)

    def Reg(self, modulename, el, args):
        width = args[0] if args is not None else 1
        symbol = Symbol(self.varname(modulename, el.name), BVType(width))
        self.add_var(modulename, el.name, symbol)
        self.reglist.append(symbol)
        return (el.name, args)
    
    def Ioport(self, modulename, el, args):
        return (el.first, args)

    def Input(self, modulename, el, args):
        return (el.name, args)

    def Inout(self, modulename, el, args):
        return (el.name, args)
    
    def Output(self, modulename, el, args):
        return (el.name, args)

    def Port(self, modulename, el, args):
        return (None, [(el.name, None)])

    def Portlist(self, modulename, el, args):
        for port in args:
            porttype = port[0]
            portname = port[1][0][0]
            portsize = port[1][0][1]

            if porttype == None:
                if portsize is None:
                    self.add_var(modulename, portname, (None, INPUT))
                else:
                    width = portsize[0]
                    var = Symbol(self.varname(modulename, portname), BVType(width))
                    self.add_var(modulename, portname, var)
                    self.ts.add_input_var(var)
                    self._add_clock_behavior(var, modulename)
                continue
            
            if type(porttype) == Output:
                if portsize is None:
                    self.add_var(modulename, portname, (None, OUTPUT))
                else:
                    width = portsize[0]
                    var = Symbol(self.varname(modulename, portname), BVType(width))
                    self.add_var(modulename, portname, var)
                    self.ts.add_output_var(var)
                    self.outputlist.append(var)
                continue

            if type(porttype) == Input:
                if portsize is None:
                    self.add_var(modulename, portname, (None, INPUT))
                else:
                    width = portsize[0]
                    var = Symbol(self.varname(modulename, portname), BVType(width))
                    self.add_var(modulename, portname, var)
                    self.ts.add_input_var(var)
                    self._add_clock_behavior(var, modulename)
                continue
                    
            assert False

        self.portslist = [p[1][0][0] for p in args]

        return args

    def add_var(self, modulename, varname, value):
        self.varmap[varname] = value

    def get_var(self, modulename, varname, size=1):
        if type(self.varmap[varname]) == tuple:
            var = Symbol(self.varname(modulename,varname), BVType(size))
            if self.varmap[varname][1] == INPUT:
                self.ts.add_input_var(var)
                self._add_clock_behavior(var, modulename)
            if self.varmap[varname][1] == OUTPUT:
                self.ts.add_output_var(var)
                self.outputlist.append(var)
            self.varmap[varname] = var
            return var
        else:
            return self.varmap[varname]
        
    def Decl(self, modulename, el, args):
        if args[0] == None:
            return None

        direction = el.children()[0]

        if type(direction) in [Input, Output, Inout, Reg, Wire]:
            width = args[0][1]
            typ = el.children()[0]
            typname = typ.name

            if self.hide_autogenerated:
                typname = args[0][0]
            
            if width is None:
                if typname not in self.varmap:
                    Logger.error("Variable \"%s\" not defined, line %d"%(typname, el.lineno))
                var_inmap = self.varmap[typname]
                if type(var_inmap) == tuple:
                    width = 1 if var_inmap[0] is None else var_inmap[0]
                else:
                    width = var_inmap.symbol_type().width
            else:
                width = width[0]
            
            if typname in self.varmap:
                prev_width = self.varmap[typname][0] if type(self.varmap[typname]) is not FNode else self.varmap[typname].symbol_type().width
                if (prev_width is not None) and (prev_width != width):
                    Logger.error("Unmatched variable size at line %d"%el.lineno)

            var = Symbol(self.varname(modulename, typname), BVType(width))

            if type(direction) == Input:
                self.ts.add_input_var(var)
                self._add_clock_behavior(var, modulename)
            elif type(direction) == Output:
                self.ts.add_output_var(var)
                self.outputlist.append(var)
            elif type(direction) == Reg:
                self.ts.add_state_var(var)
            else:
                self.ts.add_var(var)

            self.add_var(modulename, typname, var)

            return None

        if type(direction) == RegArray:
            low, high = args[0][1][1], args[0][1][0]
            low, high = min(low, high), max(low, high)
            width = args[0][0]
            vname = el.children()[0].name
            var_idxs = []
            for i in range(low, high+1, 1):
                vname_idx = self.get_num_name(vname, i, high)
                var_idx = Symbol(self.varname(modulename, vname_idx), BVType(width))
                var_idxs.append(var_idx)
                self.add_var(modulename, vname_idx, var_idx)
                self.ts.add_state_var(var_idx)
                self.memlist.append(var_idx)

            self.add_var(modulename, vname, var_idxs)
            self.ts.add_hidden_var(Symbol(self.varname(modulename, vname), BVType(width)))
                
            return None
        
        Logger.error("Unmanaged type %s, line %d"%(type(direction), el.lineno))

    def get_num_name(self, name, idx, maxval):
        return "%s_%s"%(name, str(idx).rjust(len(str(maxval)),"0"))
        
    def Sens(self, modulename, el, args):
        if el.sig is None:
            return TRUE()
            
        var = self.get_var(modulename, el.sig.name)

        zero = BV(0, var.symbol_type().width)
        one = BV(1, var.symbol_type().width)

        if el.type == POSEDGE:
            if self.config.abstract_clock:
                self.abstract_clock_list.append((var, (zero, one)))
                return EqualsOrIff(var, one)
            return And(EqualsOrIff(TS.get_prime(var), one), EqualsOrIff(var, zero))

        if el.type == NEGEDGE:
            if self.config.abstract_clock:
                Logger.warning("Clock abstraction activated with negedge behavior")
                self.abstract_clock_list.append((var, (one, zero)))
                return EqualsOrIff(var, zero)

            return And(EqualsOrIff(TS.get_prime(var), zero), EqualsOrIff(var, one))

        if el.type == LEVEL:
            return EqualsOrIff(var, one)
        
        Logger.error("Unknown driver type \"%s\" at line %d"%(el.type, el.lineno))

    def Lvalue(self, modulename, el, args):
        return args[0]

    def Rvalue(self, modulename, el, args):
        return args[0]

    def NonblockingSubstitution(self, modulename, el, args):
        left, right = args[0], args[1]

        if type(right) == int:
            right = BV(right, get_type(left).width)

        if left.is_symbol():
            left = TS.to_next(left)
        else:
            primedic = dict([(v.symbol_name(), TS.get_prime(v).symbol_name()) for v in get_free_variables(left) \
                             if v not in self.ts.input_vars])
            left = substitute(left, primedic)
            
        return Assign(left, right)

    def BlockingSubstitution(self, modulename, el, args):
        left, right = args[0], args[1]
        delay = 0
        
        if len(args) > 2:
            delay = args[2]
            
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)

        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
            
        if type(left) == int:
            left = BV(left, MAXINT)
        if type(right) == int:
            right = BV(right, MAXINT)
            
        fv_left = get_free_variables(left)
        fv_right = get_free_variables(right)

        if delay == 1:
            left = TS.to_next(left)
            
        return Define(left, B2BV(right))

    def SensList(self, modulename, el, args):
        return Or(args)

    def Integer(self, modulename, el, args):
        intvar = Symbol(self.varname(modulename, el.name), BVType(MAXINT))
        self.add_var(modulename, el.name, intvar)
        return None
    
    def IntConst(self, modulename, el, args):
        if "'d" in el.value:
            width, value = el.value.split("'d")
            if width == "":
                return int(value)
            return BV(int(value), int(width))

        if "'b" in el.value:
            width, value = el.value.split("'b")
            if value == "z":
                if width == "":
                    Logger.error("High Bit value definition requires size, line %d"%el.lineno)
                value = dec_to_bin(int((2**int(width))-1), int(width))
            if width == "":
                return int(bin_to_dec(value))

            if (type(value) == str) and ("x" in value):
                if width == "":
                    Logger.error("Non deterministic value definition requires size, line %d"%el.lineno)

                valname = DONTCARE_VAR%value
                valvar = Symbol(self.varname(modulename, valname), BVType(int(width)))
                self.add_var(modulename, valname, valvar)
                self.ts.add_var(valvar)
                # Inverting the value, e.g., with 'b1x x is the least significant
                value = value[::-1]
                if len(value) > 1:
                    valconstr = TRUE()
                    for val in value:
                        if val == "x":
                            continue
                        index = value.index(val)
                        invalue = BV(0,1) if int(val) == 0 else BV(1,1)
                        valconstr = And(valconstr, EqualsOrIff(BVExtract(valvar, index, index), invalue))
                    self.add_constraint(valconstr)
                return valvar

            return BV(int(bin_to_dec(value)), int(width))
        
        return int(el.value)

    def Identifier(self, modulename, el, args):
        elname = el.name

        if self.hide_autogenerated:        
            if self.is_autogenerated(elname):
                elname = NODE_SYMBOL%elname
        
        if elname in self.paramdic:
            return self.paramdic[elname]

        if elname in self.varmap:
            return self.get_var(modulename, elname)

        Logger.error("Symbol \"%s\" is not defined, line %d"%(elname, el.lineno))
        return elname
    
    def Width(self, modulename, el, args):
        return (args[0] - args[1])+1
    
    def Block(self, modulename, el, args):
        return [a for a in args if a is not None]

    def Cond(self, modulename, el, args):
        if args[0] in [False, FALSE()]:
            return args[2]
        if args[0] in [True, TRUE()]:
            return args[1]
        return Ite(BV2B(args[0]), args[1], args[2])
    
    def IfStatement(self, modulename, el, args):
        statement, then_b, else_b = args[0], args[1], args[2] if len(args) > 2 else None

        if type(then_b) == list:
            if len(then_b) == 0:
                then_b = None
            elif type(then_b[0]) == FNode:
                then_b = And(then_b)
            elif type(then_b[0]) == ProcessedAlways:
                pass
            else:
                Logger.error("Not Implemented")

        if type(else_b) == list:
            if len(else_b) == 0:
                else_b = None
            elif type(else_b[0]) == FNode:
                else_b = And(else_b)
            elif type(else_b[0]) == ProcessedAlways:
                pass
            else:
                Logger.error("Not Implemented")
                
        if type(statement) == int:
            condition = self.to_bool(statement)
        elif (type(statement) == bool) or (get_type(statement) == BOOL):
            condition = statement
        else:
            one = BV(1, get_type(statement).width)
            condition = EqualsOrIff(statement, one)

        if else_b is None:
            if then_b is None:
                return None
            if condition in [True, TRUE()]:
                return then_b
            if condition in [False, FALSE()]:
                return None
            return Implies(condition, then_b)
        else:
            if condition in [True, TRUE()]:
                assert (then_b is not None)
                return then_b
            if condition in [False, FALSE()]:
                assert (else_b is not None)
                return else_b
            assert (then_b is not None) and (else_b is not None)
            return Ite(condition, then_b, else_b)

    def Initial(self, modulename, el, args):
        self.add_init(And([And(a) for a in args]))
        return And(args)

    def process_always(self, modulename, always_list):
        if len(always_list) == 0:
            return

        g_always = {}

        # Combine all conditions
        for always in always_list:
            for v in always:
                if v not in g_always:
                    g_always[v] = []
                g_always[v] += always[v]

        state_vars = set(self.reglist+self.memlist)
        possible_ass_vars = set(self.reglist+self.outputlist+self.memlist)
        ass_vars = set(g_always.keys())
        diff_vars = ass_vars.difference(possible_ass_vars)

        strip_name = modulename[:modulename.find(MODPARST)] if MODPARST in modulename else modulename
        
        if len(diff_vars) > 0:
            Logger.error("Assignment is not allowed to variables \"%s\" in module \"%s\""%(", ".join([v.symbol_name().replace("%s."%(modulename),"") for v in diff_vars]), strip_name))

        ftrans = {}

        for var in possible_ass_vars:
            keep_value = EqualsOrIff(var, TS.get_prime(var))
            if (var not in g_always):
                if (var in state_vars):
                    self.add_constraint(keep_value)
                continue

            for beh in g_always[var]:
                effect, condition = self.assign_define_substitute(beh[0])[0], beh[1]
                condition = simplify(condition)

                # assignment and value from the assign operator (this can differ from var with the addition of bitvector extraction)
                avar, aval = beh[0].args()[0], beh[0].args()[1]
                # Only assignments to primed variables are added to the functional transition relation
                cons = simplify(Implies(condition, effect))

                if avar not in ftrans:
                    ftrans[avar] = []
                ftrans[avar].append((condition, aval))

        # Sort and simplifies assignments in case of ITE encoding
        sort_assignments = False
        
        for var, asscond in ftrans.items():
            if not sort_assignments:
                self.add_ftrans(var, asscond)
            else:
                sim_asscond = [(len(list(conjunctive_partition(x[0]))), str(x[1]), list(conjunctive_partition(x[0])), x[1]) for x in asscond]
                sim_asscond.sort()
                cum_cond = []
                new_assconds = []
                for el in sim_asscond:
                    cp_cond, value = el[2], el[3]
                    new_asscond = And([x for x in cp_cond if x not in cum_cond])
                    cum_cond += [Not(x) for x in cp_cond]
                    new_assconds.append((new_asscond, value))
                self.add_ftrans(var, new_assconds)

    # Simplifies ites according with a provided condition
    def simplify_ites(self, formula, condition):
        subwalker = IteSubstituteWalker()
        subwalker.set_condition(condition)
        return subwalker.walk(formula)
            
    def Always(self, modulename, el, args):
        condition, statements = args[0], args[1]

        if type(statements) == FNode:
            statements = [statements]

        if And(statements) == TRUE():
            return TRUE()
            
        # All variables in the senselist are accessed at the primed time
        sensedict = dict([(v.symbol_name(), TS.get_prime(v).symbol_name()) \
                          for v in get_free_variables(condition) if not TS.is_prime(v)])
        
        for i in range(len(statements)):
            statements[i] = substitute(statements[i], sensedict)

        assign_conditions = self.collect_assign_conditions(And(statements))

        # The variable assignments apply only when the "always condition" is true
        for v in assign_conditions:
            for i in range(len(assign_conditions[v])):
                avalue, acond = assign_conditions[v][i][0], assign_conditions[v][i][1]
                avalue = self.simplify_ites(avalue, acond)
                assign_conditions[v][i] = (avalue, And(acond, condition))
            
        # Assignlist is a list of (variable, list((value, condition))), meaning that
        # the "variable" is assigned to "value" when "condition" is true
                
        # The result is the processed by the process_always function
        return ProcessedAlways(assign_conditions)
    
    def ForStatement(self, modulename, el, args):

        # Refining definition in primed format e.g., for(i=0;i<10;i=i+1) into for(i'=0;i<10;i'=i+1)
        args[0] = args[0].substitute(dict([(v, TS.get_prime(v)) for v in get_free_variables(args[0])]))
        cp = list(conjunctive_partition(args[2]))
        newcp = []
        for ass in cp:
            left, right = ass.args()[0], ass.args()[1]
            assert (ass.is_equals() or ass.is_iff() or \
                    ass.node_type() == ASSIGN or \
                    ass.node_type() == DEFINE) and \
                    (left.is_symbol() or right.is_symbol())
            if left.is_symbol():
                newcp.append(EqualsOrIff(TS.get_prime(left), right))
            if right.is_symbol():
                newcp.append(EqualsOrIff(TS.get_prime(right), left))

        args[2] = And(newcp)

        args_0 = self.assign_define_substitute(args[0])[0]
        args_1 = self.assign_define_substitute(args[1])[0]
        args_2 = self.assign_define_substitute(args[2])[0]
        
        # Performining the For-loop unrolling
        fv = get_free_variables(args_0)
        model = get_model(args_0)
        state_n = [(v, model[v]) for v in fv]
        state_c = [(TS.get_ref_var(v), model[v]) for v in fv]
        state = dict(state_c + state_n)
        formulae = []
        while True:
            # Evaluate new instance
            formula = simplify(And(args[3]).substitute(state))
            formulae.append(formula)

            # Compute next step
            model = get_model(And(And([EqualsOrIff(v[0], v[1]) for v in state_c]), args_2))
            state_n = [(v, model[v]) for v in fv]
            state_c = [(TS.get_ref_var(v), model[v]) for v in fv]
            state = dict(state_c + state_n)

            # Exit condition
            if not is_sat(And(And([EqualsOrIff(v[0], v[1]) for v in state_c]), args_1)):
                break

        return And(formulae)

    def ModuleDef(self, modulename, el, args):
        always_list = [a.assign_conditions for a in args if type(a) == ProcessedAlways]
        for a in args:
            if (type(a) == list) and (len(a) > 0) and (type(a[0]) == ProcessedAlways):
                always_list += [x.assign_conditions for x in a]
        self.process_always(modulename, always_list)

        self.hts.add_ts(self.ts)

        self.portslist.sort()
        for param in self.portslist:
            self.hts.add_param(self.get_var(modulename, param))

        Logger.log("Ports module \"%s\": %s"%(el.name, ", ".join(["%s:%s"%(p, get_type(self.get_var(modulename, p))) for p in self.portslist])), 2)
        Logger.log("Parameters module \"%s\": %s"%(el.name, ", ".join(["%s=%s"%(p, self.paramdic[p]) for p in self.paramdic])), 2)
        return self.hts

    def Description(self, modulename, el, args):
        return args[0]

    def Source(self, modulename, el, args):
        return args[0]

    def Divide(self, modulename, el, args):
        left, right = args[0], args[1]

        if (type(left) == int) and (type(right) == int):
            return int(left/right)
        
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)

        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
            
        if (type(right) == int):
            right = BV(right, MAXINT)

        if (type(left) == int):
            left = BV(left, MAXINT)
            
        return BVUDiv(left, right)

    def Uminus(self, modulename, el, args):
        return self.Minus(modulename, el, [0, args[0]])
    
    def Minus(self, modulename, el, args):
        left, right = args[0], args[1]

        if (type(left) == int) and (type(right) == int):
            return left-right
        
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)

        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
            
        if (type(right) == int):
            right = BV(right, MAXINT)

        if (type(left) == int):
            left = BV(left, MAXINT)
            
        return BVSub(left, right)

    def Plus(self, modulename, el, args):
        left, right = args[0], args[1]

        if (type(left) == int) and (type(right) == int):
            return left+right
        
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)

        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
            
        if (type(right) == int):
            right = BV(right, MAXINT)

        if (type(left) == int):
            left = BV(left, MAXINT)

        lft_width = get_type(left).width
        rgt_width = get_type(right).width

        if lft_width != rgt_width:
            if lft_width < rgt_width:
                left = BVConcat(left, BV(0, rgt_width-lft_width))
            if lft_width > rgt_width:
                right = BVConcat(right, BV(0, lft_width-rgt_width))

        return BVAdd(left, right)

    def Times(self, modulename, el, args):
        left, right = args[0], args[1]

        if (type(left) == int) and (type(right) == int):
            return left*right
        
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)

        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
            
        if (type(right) == int):
            right = BV(right, MAXINT)

        if (type(left) == int):
            left = BV(left, MAXINT)

        return BVMul(left, right)
    
    def Eq(self, modulename, el, args):
        left, right = args[0], args[1]
        
        if left == right:
            return TRUE()

        if (type(left) == int) and (type(right) == int):
            return left==right        
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
        
        return EqualsOrIff(left, right)

    def Eql(self, modulename, el, args):
        return self.Eq(modulename, el, args)
    
    def NotEq(self, modulename, el, args):
        return Not(self.Eq(modulename, el, args))

    def NotEql(self, modulename, el, args):
        return self.NotEq(modulename, el, args)
    
    def to_bool(self, el):
        if type(el) == int:
            return FALSE() if el == 0 else TRUE()
        if type(el) == bool:
            return TRUE() if el else FALSE()
        return BV2B(el)

    def to_bv(self, el):
        if type(el) == int:
            return BV(el, 1)
        return B2BV(el)

    def Concat(self, modulename, el, args):
        return self._multi_op(BVConcat, args)
    
    def And(self, modulename, el, args):
        assert len(args) == 2
        return BVAnd(self.to_bv(args[0]), self.to_bv(args[1]))
    
    def Xor(self, modulename, el, args):
        assert len(args) == 2
        return BVXor(self.to_bv(args[0]), self.to_bv(args[1]))

    def Or(self, modulename, el, args):
        assert len(args) == 2
        return BVOr(self.to_bv(args[0]), self.to_bv(args[1]))

    def Uor(self, modulename, el, args):
        assert len(args) == 1
        # Or reduce
        zero = BV(0, get_type(args[0]).width)
        return Ite(EqualsOrIff(args[0], zero), BV(0,1), BV(1,1))

    def Uand(self, modulename, el, args):
        assert len(args) == 1
        # And reduce
        width = get_type(args[0]).width
        ones = BV((2**width)-1, width)
        return Ite(EqualsOrIff(args[0], ones), BV(1,1), BV(0,1))


    def Land(self, modulename, el, args):
        if set([type(a) for a in args]) == set([bool]):
            ret = args[0]
            for a in args[1:]:
                ret = ret or a
            return ret
        return simplify(And([self.to_bool(x) for x in args]))
    
    def Lor(self, modulename, el, args):
        if set([type(a) for a in args]) == set([bool]):
            ret = args[0]
            for a in args[1:]:
                ret = ret or a
            return ret
        return simplify(Or([self.to_bool(x) for x in args]))
    
    def Sll(self, modulename, el, args):
        left, right = args[0], args[1]
        if (type(left) == int) and (type(right) == int):
            return left << right
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
        
        return BVLShl(left, right)
        
    def LessThan(self, modulename, el, args):
        left, right = args[0], args[1]

        if (left == None) or (right == None):
            return False
        
        if (type(left) == int) and (type(right) == int):
            return left < right
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
            
        return BVULT(left, right)

    def LessEq(self, modulename, el, args):
        left, right = args[0], args[1]

        if (left == None) or (right == None):
            return False
        
        if (type(left) == int) and (type(right) == int):
            return left <= right
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
            
        return BVULE(left, right)
    
    def GreaterThan(self, modulename, el, args):
        left, right = args[0], args[1]

        if (left == None) or (right == None):
            return False
        
        if (type(left) == int) and (type(right) == int):
            return left > right
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
            
        return BVUGT(left, right)

    def GreaterEq(self, modulename, el, args):
        left, right = args[0], args[1]

        if (left == None) or (right == None):
            return False
        
        if (type(left) == int) and (type(right) == int):
            return left >= right
        if (type(left) == int) and (type(right) == FNode):
            left = BV(left, get_type(right).width)
        if (type(right) == int) and (type(left) == FNode):
            right = BV(right, get_type(left).width)
            
        return BVUGE(left, right)
    
    def Ulnot(self, modulename, el, args):
        return BVNot(args[0])

    def Unot(self, modulename, el, args):
        if type(args[0]) == int:
            return Not(self.to_bool(args[0]))
        zero = BV(0, get_type(args[0]).width)
        return EqualsOrIff(args[0], zero)
    
    def Parameter(self, modulename, el, args):
        if el.name not in self.paramdic:
            self.paramdic[el.name] = args[0]
        return None

    def SingleStatement(self, modulename, el, args):
        return None
    
    def SystemCall(self, modulename, el, args):
        if el.syscall == "clog2":
            return math.ceil(math.log(args[0])/math.log(2))

        if el.syscall == "display":
            return None

        if el.syscall == "finish":
            return None

        if el.syscall == "time":
            return None
        
        Logger.error("Unimplemented system call \"%s\", line %d"%(el.syscall, el.lineno))
        
    def Partselect(self, modulename, el, args):
        if abs(args[1]-args[2])+1 == get_type(args[0]).width:
            return args[0]
        return BVExtract(args[0], args[2], args[1])

    def Assign(self, modulename, el, args):
        left, right = args[0], args[1]

        if type(right) == int:
            right = BV(right, get_type(left).width)

        if get_type(left).is_bv_type() and get_type(right).is_bv_type():
            
            lft_width = get_type(left).width
            rgt_width = get_type(right).width

            if lft_width == rgt_width + 1:
                fv = get_free_variables(right)
                right = right.substitute(dict([(v, BVConcat(v, BV(0, 1))) for v in fv]))

            lft_width = get_type(left).width
            rgt_width = get_type(right).width

            if lft_width > rgt_width:
                right = BVConcat(right, BV(0, lft_width-rgt_width))

        invar = EqualsOrIff(B2BV(left), B2BV(right))
        self.add_invar(invar)
        return el
    
    def Pointer(self, modulename, el, args):
        if type(args[0]) == tuple:
            width = get_type(args[1]).width
            mem_size = args[0][1]
            mem_locations = [self.get_num_name(args[0][0], i, mem_size[1]+1) for i in range(mem_size[0], mem_size[1]+1)]
            ma = mem_access(args[1], [self.varmap[v] for v in mem_locations], width)
            return ma

        if (type(args[0]) == FNode) and (type(args[1]) == FNode):
            width_mem = get_type(args[0]).width
            width_idx = get_type(args[1]).width
            mem_locations = [BVExtract(args[0], i, i) for i in range(width_mem)]
            ma = mem_access(args[1], mem_locations, width_idx)
            return ma

        # Management of memory access
        if (type(args[0]) == list) and (type(args[1]) == FNode):
            width_mem = len(args[0])
            width_idx = get_type(args[1]).width
            mem_locations = args[0]
            ma = mem_access(args[1], mem_locations, width_idx)
            return ma

        if (type(args[0]) == list) and (type(args[1]) == int):
            return args[0][args[1]]
        
        return BVExtract(args[0], args[1], args[1])

    def LConcat(self, modulename, el, args):
        if len(args) == 1:
            return args[0]
        assert len(args) == 2
        return BVConcat(args[0], args[1])
    
    def _rec_repeat(self, el, count):
        if count == 1:
            return el
        return simplify(BVConcat(el, self._rec_repeat(el, count-1)))

    def _multi_op(self, op, args):
        if len(args) == 1:
            return args[0]
        ret = op(args[0], args[1])
        if len(args) > 2:
            for el in args[2:]:
                ret = op(ret, el)
        return ret
    
    def Repeat(self, modulename, el, args):
        return self._rec_repeat(args[0], args[1])
    
    def PortArg(self, modulename, el, args):
        return (el.portname, args[0])

    def Instance(self, modulename, el, args):
        el_child = el.children()
        paramargs_idx = [el_child.index(p) for p in el_child if type(p) == ParamArg]
        portargs_idx = [el_child.index(p) for p in el_child if type(p) == PortArg]
        width_idx = [el_child.index(p) for p in el_child if type(p) == Width]

        portargs = [args[i] for i in portargs_idx]
        if len([p[0] for p in portargs if p[0] is None]) == 0:
            portargs.sort()
        paramargs = [args[i] for i in paramargs_idx]
        paramargs.sort()
        width = args[width_idx[0]] if len(width_idx) > 0 else None

        if el.module in SV_MODULES:
            if el.module == SVA_ASSERT:
                sname = ASSERT_SYMBOL%el.lineno
                asymbol = Symbol(self.varname(modulename, sname), BOOL)
                self.add_var(modulename, sname, asymbol)
                self.ts.add_var(asymbol)
                self.add_constraint(EqualsOrIff(asymbol, And([BV2B(a[1]) for a in args])))
            if el.module == SVA_AT:
                out = dict(args)["o"]
                in0 = dict(args)["a0"]
                in1 = dict(args)["a1"]
                self.add_constraint(EqualsOrIff(Implies(BV2B(in0), BV2B(in1)), BV2B(out)))
            if el.module == SVA_POSEDGE:
                out = dict(args)["o"]
                in_ = dict(args)["i"]
                if self.config.abstract_clock:
                    self.add_constraint(EqualsOrIff(BV2B(out), EqualsOrIff(in_, BV(1, 1))))
                else:
                    self.add_constraint(EqualsOrIff(BV2B(TS.to_next(out)), And(EqualsOrIff(in_, BV(0, 1)), EqualsOrIff(TS.to_next(in_), BV(1, 1)))))
            if el.module == SVA_IMMEDIATE_ASSERT:
                sname = ASSERTI_SYMBOL%el.lineno
                asymbol = Symbol(self.varname(modulename, sname), BOOL)
                self.add_var(modulename, sname, asymbol)
                self.ts.add_var(asymbol)
                self.add_constraint(EqualsOrIff(asymbol, And([BV2B(a[1]) for a in args])))
                
            return args
        
        if el.module not in self.modulesdic:
            if el.module in self.mod_map:
                actual_args = [p[1] for p in portargs]
                ts = self.mod_map[el.module](actual_args[0], *(actual_args[1:]))
                self.hts.add_ts(ts)
                return args
            else:
                Logger.error("Module definition \"%s\" not found, line %d"%(el.module, el.lineno))
        
        param_modulename = self.clean_name(el.module)

        include_varargs = False
        varmap = {}
        
        parameterized_type = ["%s%s"%(p[0], p[1]) for p in paramargs]

        if include_varargs:
            parameterized_type += ["%s%s"%(p[0], get_type(p[1]).width) for p in portargs]
            for portarg in portargs:
                varmap[portarg[0]] = (get_type(portarg[1]).width, None)

        hide_sub_vars = False
                
        if self.hide_autogenerated:
            if self.is_autogenerated(param_modulename):
                hide_sub_vars = True
                
        if len(parameterized_type) > 0:
            param_modulename = "".join((param_modulename, MODPARST, ",".join(parameterized_type), MODPAREN))

        instances = []

        # Management of multi instances e.g., type instance[k:0] (clk, input[k:0])
        if len(width_idx) > 0:
            formal_args = [p[0] for p in portargs]
            args_width_idx = [formal_args.index(c.portname) for c in el_child if type(c.children()[0]) == Partselect]
            map_ports = dict([(dict(portargs)[c.portname], dict(portargs)[c.portname].args()[0]) for c in el_child if type(c.children()[0]) == Partselect and dict(portargs)[c.portname].is_bv_extract()])
            remap_ports = [(p[0], map_ports[p[1]] if p[1] in map_ports else p[1]) for p in portargs]

            extract_port = lambda p,i: BVExtract(p, i, i)
            
            for i in range(args[width_idx[0]]):
                idx_params = [(p[0], extract_port(p[1], i) if remap_ports.index(p) in args_width_idx else p[1]) for p in remap_ports]
                idx_instance_name = self.get_num_name(el.name, i, args[width_idx[0]])
                instances.append((idx_instance_name, idx_params))
        else:        
            instances = [(el.name, portargs)]

        for (instance, actualargs) in instances:
            instance = self.clean_name(instance)
            instancewalker = VerilogSTSWalker()
            instancewalker.config = copy.deepcopy(self.config)
            instancewalker.config.add_clock = False
            instancewalker.paramdic = dict(paramargs)
            instancewalker.varmap = varmap
            instancewalker.modulesdic = self.modulesdic
            subhts = instancewalker.walk_module(self.modulesdic[el.module], param_modulename)
            subhts.name = param_modulename
                
            # Setting parameters to value 0 in case they are not provided
            if len(subhts.params) != len(actualargs):
                dic_form = dict([(".".join(p.symbol_name().split(".")[1:]), p) for p in subhts.params])
                lst_actu = [str(a[0]) for a in actualargs]

                for par in dic_form:
                    if par not in lst_actu:
                        actualargs.append((par, BV(0, dic_form[par].symbol_type().width)))
                actualargs.sort()

            if hide_sub_vars:
                subhts.apply_var_prefix(HIDDEN_VAR)
                
            formal_ptype = [get_type(p) for p in subhts.params]
            actual_ptype = [get_type(a[1]) for a in actualargs]
            if formal_ptype != actual_ptype:
                formal = ["%s:%s"%(".".join(p.symbol_name().split(".")[1:]), get_type(p)) for p in subhts.params]
                actual = ["%s:%s"%(a[0], get_type(a[1])) for a in actualargs]
                Logger.error("Parameters type error, line %d\nformal=\"%s\" \nactual=\"%s\""%(el.lineno, formal, actual))
            self.hts.add_sub(instance, subhts, tuple([a[1] for a in actualargs]))
        return args

    def InstanceList(self, modulename, el, args):
        return args

    def ParamArg(self, modulename, el, args):
        return (el.paramname, args[0])

    def StringConst(self, modulename, el, args):
        return el.value

    def Localparam(self, modulename, el, args):
        return self.Parameter(modulename, el, args)

    def GenerateStatement(self, modulename, el, args):
        if len(args) == 1:
            return args[0]
        return args

    def Length(self, modulename, el, args):
        return args

    def RegArray(self, modulename, el, args):
        return args

    def IdentifierScope(self, modulename, el, args):
        return el
    
    def IdentifierScopeLabel(self, modulename, el, args):
        Logger.error("Unsupported indentifier scope, line %d"%(el.lineno))
        return el

    def DelayStatement(self, modulename, el, args):
        delay = int(el.delay.value)
        if delay == 1:
            return delay
        Logger.error("Unsupported delay statement != 1, line %d"%(el.lineno))
        return el

    def ForeverStatement(self, modulename, el, args):
        statement = And(args)
        for el in [And(a) for a in args]:
            self.add_constraint(el)
        return TRUE()

    def collect_conditions_rec(self, formula, ext_function, assign_list=[], collected={}):

        if formula.is_ite():
            listlen = len(assign_list)
            condition, then_b, else_b = formula.args()[0], formula.args()[1], formula.args()[2]
            (assign_list, collected) = self.collect_conditions_rec(then_b, ext_function, \
                                                                   assign_list[:listlen]+[condition], collected)
            (assign_list, collected) = self.collect_conditions_rec(else_b, ext_function, \
                                                                   assign_list[:listlen]+[Not(condition)], collected)

        if formula.is_implies():
            condition, then_b = formula.args()[0], formula.args()[1]
            (assign_list, collected) = self.collect_conditions_rec(then_b, ext_function, \
                                                                   assign_list+[condition], collected)

        if formula.is_and():
            lstlen = len(assign_list)

            for arg in formula.args():
                (assign_list, collected) = self.collect_conditions_rec(arg, ext_function, assign_list[:lstlen], collected)

        (assign_list, collected) = ext_function(formula, assign_list, collected)

        return (assign_list, collected)

    def collect_assign_conditions(self, formula):

        # Reduces an assignment (var = 0) & (var != 1) & (var != 2) .. into (var = 0)
        def simplify_assign_list(assign_list):
            cp = list(conjunctive_partition(assign_list))
            equals = [f for f in cp if f.is_equals()]
            if len(equals) != 1:
                if cp[0].args():
                    # Managing condition where all assignments are negated
                    get_eq_val = lambda f: f.args()[0] if f.args()[0].is_bv_constant() else f.args()[1]
                    get_eq_var = lambda f: f.args()[0] if not f.args()[0].is_bv_constant() else f.args()[1]
                    values = [get_eq_val(f.args()[0]).constant_value() for f in cp]
                    if len(values) == len(range(max(values)+1)):
                        refvar = get_eq_var(cp[0].args()[0])
                        return BVUGT(refvar, BV(max(values), refvar.symbol_type().width))
                
                return assign_list

            refel = equals[0]
            left, right = refel.args()[0], refel.args()[1]
            pos_var = left if left.is_symbol() else right
            pos_val = right if left.is_symbol() else left
            
            ret = [refel]
            for neq in [f.args()[0] for f in cp if f.is_not()]:
                if not neq.is_equals():
                    return assign_list

                left, right = neq.args()[0], neq.args()[1]
                neg_var = left if left.is_symbol() else right
                neg_val = right if left.is_symbol() else left

                if not((pos_var == neg_var) & (pos_val != neg_val)):
                    ret.append(neq)

            return And(ret)

        # External function for collect_conditions_rec. Symbols collection
        def var_fun(formula, assign_list, collected):
            if formula.is_symbol():
                ref_var = TS.get_ref_var(formula)
                if ref_var not in collected:
                    collected[ref_var] = []
                collected[ref_var].append((formula, And(assign_list)))
            return (assign_list, collected)
        
        # External function for collect_conditions_rec. Assigns and Defines collection
        def assign_define_fun(formula, assign_list, collected):
            left = formula.args()[0]
            if formula.node_type() in [ASSIGN, DEFINE]:
                if not left.is_symbol():
                    fv = get_free_variables(left)
                    mem_vars = [TS.get_ref_var(v) for v in fv if TS.get_ref_var(v) in self.memlist]
                    idx_vars = [TS.get_ref_var(v) for v in fv if TS.get_ref_var(v) not in self.memlist]

                    if left.is_bv_extract():
                        left = left.args()[0]
                    
                    var_conditions = self.collect_conditions_rec(left, var_fun, assign_list=[], collected={})[1]

                    lstlen = len(assign_list)
                    for var, conditions in var_conditions.items():
                        for cond in conditions:
                            assign_list = assign_list[:lstlen]
                            newcond = And(cond[1])
                            newcond = simplify_assign_list(newcond)
                            assign_list.append(newcond)
                            if var not in collected:
                                collected[var] = []
                            collected[var].append((formula, And(assign_list)))
                else:
                    ref_var = TS.get_ref_var(left)
                    if ref_var not in collected:
                        collected[ref_var] = []
                    collected[ref_var].append((formula, And(assign_list)))

            return (assign_list, collected)

        (assign_list, collected) = self.collect_conditions_rec(formula, assign_define_fun, assign_list=[], collected={})
        
        return collected

class IteSubstituteWalker(IdentityDagWalker):

    condition = None

    def set_condition(self, condition):
        self.condition = condition

    def walk_ite(self, formula, args, **kwargs):
        if is_unsat(And(args[0], self.condition)):
            return args[2]

        if is_unsat(And(Not(args[0]), self.condition)):
            return args[1]

        return formula
        
class AssignDefineSubstituteWalker(IdentityDagWalker):

    assign_list = None

    def reset(self):
        self.assign_list = set([])
    
    def walk_assign(self, formula, args, **kwargs):
        for v in get_free_variables(args[0]):
            self.assign_list.add(v)
        return self.mgr.EqualsOrIff(args[0], args[1])

    def walk_define(self, formula, args, **kwargs):
        for v in get_free_variables(args[0]):
            self.assign_list.add(v)
        return self.mgr.EqualsOrIff(args[0], args[1])
    
class ProcessedAlways(object):
    def __init__(self, assign_conditions):
        self.assign_conditions = assign_conditions



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
import inspect

ERROR = "ERROR: "

class Logger(object):
    verbosity = 0
    linenum_verbosity = 5
    id_timer = 0
    timers = []
    time = False
    single_warnings = True
    prev_warnings = None
    error_raise_exept = True
    newline = True

    _last_inline = None
    
    @staticmethod        
    def msg(msg, level, condition=True, max_level=10):
        if Logger.verbosity > 1:
            Logger.log(msg, level, condition, max_level)
        else:
            if (Logger.verbosity > level) and (Logger.verbosity <= max_level+1) and (condition):
                sys.stdout.write(msg)
                sys.stdout.flush()
                Logger.newline = "\n" in msg

    @staticmethod        
    def inline(msg, level, condition=True, max_level=10):
        if (Logger.verbosity > level) and (Logger.verbosity <= max_level+1) and (condition):
            Logger.clear_inline(0)
            lmsg = len(msg)
            Logger._last_inline = lmsg
            sys.stdout.write(msg)
            sys.stdout.write('\b'*lmsg)
            sys.stdout.flush()
            Logger.newline = "\n" in msg

    @staticmethod        
    def clear_inline(level, condition=True, max_level=10):
        if Logger._last_inline is not None:
            if (Logger.verbosity > level) and (Logger.verbosity <= max_level+1) and (condition):
                sys.stdout.write(' '*Logger._last_inline)
                sys.stdout.write('\b'*Logger._last_inline)
                sys.stdout.flush()
                Logger._last_inline = None
                Logger.newline = True

    @staticmethod        
    def line_number():
        previous_frame = inspect.currentframe().f_back.f_back
        (filename, line_number, _, _, _) = inspect.getframeinfo(previous_frame)
        return (filename, line_number)
                
    @staticmethod        
    def log(msg, level, condition=True, max_level=10):
        if (Logger.verbosity > level) and (Logger.verbosity <= max_level+1) and (condition):
            if Logger.verbosity >= Logger.linenum_verbosity:
                filename, line_number = Logger.line_number()
                msg = "\"%s\"\nfrom %s:%d"%(msg, filename, line_number)
                
            sys.stdout.write(msg+"\n")
            sys.stdout.flush()
            Logger.newline = True

    @staticmethod        
    def error(msg, raise_exception=True):
        if not Logger.newline:
            sys.stderr.write("\n")
            Logger.newline = True
        if not (Logger.error_raise_exept and raise_exception):
            sys.stderr.write("%s%s\n"%(ERROR, msg))
            sys.stderr.flush()
            sys.exit(1)
            
        raise RuntimeError(msg)

    @staticmethod        
    def warning(msg):
        if not Logger.newline:
            sys.stderr.write("\n")
            Logger.newline = True
            
        if Logger.single_warnings:
            if Logger.prev_warnings is None:
                Logger.prev_warnings = []

            if msg not in Logger.prev_warnings:
                sys.stderr.write("WARNING: "+msg+"\n")
                sys.stderr.flush()
                Logger.prev_warnings.append(msg)
        else:
            sys.stderr.write("WARNING: "+msg+"\n")
            sys.stderr.flush()
        
    @staticmethod        
    def level(level):
        return Logger.verbosity > level

    @staticmethod        
    def start_timer(name, print_time=True):
        if not Logger.time:
            return None
        if print_time:
            sys.stdout.write("Timer \"%s\": start\n"%(name))
            sys.stdout.flush()
        Logger.timers.append((time.time(), name))
        Logger.id_timer += 1
        return Logger.id_timer-1

    @staticmethod        
    def get_timer(id_timer, print_time=True):
        if not Logger.time:
            return None
        diff = time.time() - Logger.timers[id_timer][0]
        if print_time:
            sys.stdout.write("Timer \"%s\": %.2f sec\n"%(Logger.timers[id_timer][1], diff))
            sys.stdout.flush()
        return diff



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosa.encoders.factory import GeneratorsFactory, ClockBehaviorsFactory
from cosa.utils.logger import Logger

PAR_SP = ","
MODEL_SP = ";"

class ParametricBehavior(object):

    @staticmethod
    def apply_to_problem(problem, model_info):
        varsdict = dict([(var.symbol_name(), var) for var in problem.hts.vars])
        
        if problem.generators is not None:

            for strgenerator in problem.generators.split(MODEL_SP):
                strgenerator = strgenerator.replace(" ","")
                if strgenerator == "":
                    continue

                eqpos = strgenerator.find("=")
                parstart = strgenerator.find("(")
                if (parstart < eqpos) or (eqpos == -1):
                    Logger.error("Invalid generators")

                instance = strgenerator[:eqpos:]
                mdef = strgenerator[eqpos+1:]
                mtype = mdef.split("(")[0]
                pars = mdef[mdef.find("(")+1:-1].split(PAR_SP)
                generator = GeneratorsFactory.generator_by_name(mtype)
                pars = [varsdict[v] if v in varsdict else v for v in pars]
                ts = generator.get_sts(instance, pars)

                problem.hts.add_ts(ts)

        if problem.add_clock and (problem.clock_behaviors is None):
            clk_behs = []

            for (clock, (before, after)) in model_info.abstract_clock_list:
                if (clock not in model_info.clock_list) or (clock in clk_behs):
                    continue
                clock_behavior = ClockBehaviorsFactory.get_default_abstract()
                ts = clock_behavior.get_default([clock, after])
                clk_behs.append(clock)
                problem.hts.add_ts(ts)

            clock_list = [c for c in model_info.clock_list if c not in clk_behs]
            for clock in clock_list:
                if len(clock_list) > 1:
                    clock_behavior = ClockBehaviorsFactory.get_default_multi()
                else:
                    clock_behavior = ClockBehaviorsFactory.get_default()
                ts = clock_behavior.get_default([clock])
                clk_behs.append(clock)
                problem.hts.add_ts(ts)

            assert len(clk_behs) == len(set(clk_behs))

        if problem.clock_behaviors is not None:

            for strcb in problem.clock_behaviors.split(MODEL_SP):
                strcb = strcb.replace(" ","")

                if strcb == "":
                    continue

                parstart = strcb.find("(")
                parend = strcb.find(")")

                if (parstart == -1) or (parend == -1) or (parstart > parend):
                    Logger.error("Invalid Clock Behavior definition")
                
                cbname = strcb[:parstart]
                pars = strcb[parstart+1:parend].split(PAR_SP)
                pars = [varsdict[v] if v in varsdict else v for v in pars]

                clock_behavior = ClockBehaviorsFactory.clockbehavior_by_name(cbname)
                ts = clock_behavior.get_sts(pars)

                problem.hts.add_ts(ts)
        



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import cStringIO

from pysmt.shortcuts import BV, And, Or, Solver, TRUE, FALSE, Not, EqualsOrIff, Implies, Iff, Symbol, BOOL, simplify, BVAdd, BVUGE
from pysmt.rewritings import conjunctive_partition
from pysmt.smtlib.printers import SmtPrinter, SmtDagPrinter
from pysmt.logics import QF_ABV

from cosa.utils.logger import Logger
from cosa.representation import TS, HTS
from cosa.utils.formula_mngm import substitute, get_free_variables
from cosa.printers.trace import TextTracePrinter, VCDTracePrinter
from cosa.problem import Trace


class VerificationStrategy(object):
    FWD = "FWD"
    BWD = "BWD"
    ZZ  = "ZZ"
    NU  = "NU"
    INT  = "INT"
    LTL  = "LTL"
    AUTO = "AUTO"
    ALL = "ALL"

class MCConfig(object):

    incremental = True
    strategy = None
    solver = None
    prefix = None
    smt2file = None
    simplify = False
    solver_name = None
    prove = None

    def __init__(self):
        self.incremental = True
        self.strategy = VerificationStrategy.AUTO
        self.solver_name = "msat"
        self.prefix = None
        self.smt2file = None
        self.simplify = False
        self.prove = False

        self.strategies = MCConfig.get_strategies()

    @staticmethod
    def get_strategies():
        strategies = []
        strategies.append((VerificationStrategy.AUTO, "Automatic selection"))
        strategies.append((VerificationStrategy.FWD,  "Forward reachability"))
        strategies.append((VerificationStrategy.BWD,  "Backward reachability"))
        strategies.append((VerificationStrategy.ZZ,   "Mixed Forward and Backward reachability (Zig-Zag)"))
        strategies.append((VerificationStrategy.INT,  "Interpolation (not incremental only)"))
        strategies.append((VerificationStrategy.NU,   "States picking without unrolling (only for simulation)"))
        strategies.append((VerificationStrategy.LTL,  "Pure LTL verification (without optimizations)"))
        strategies.append((VerificationStrategy.ALL,  "Use all techniques"))

        return strategies

class TraceSolver(object):

    solver_name = None
    name = None
    basename = None
    trace_file = None
    solver = None
    smt2vars = None
    smt2vars_inc = None
    
    def __init__(self, solver_name, name, basename=None):
        self.solver_name = solver_name
        self.name = name
        self.basename = basename
        self.smt2vars = set([])
        self.solver = Solver(name=solver_name, logic=QF_ABV)
        self.smt2vars_inc = []
        if basename is not None:
            self.trace_file = "%s-%s.smt2"%(basename, name)

    def clear(self):
        self.solver.exit()
        self.solver = Solver(name=self.solver_name, logic=QF_ABV)

    def copy(self, name=None):
        return TraceSolver(self.solver_name, self.name if name is None else name, self.basename)
        
class BMCSolver(object):

    def __init__(self, hts, config):
        self.hts = hts
        self.config = config

        self.assert_property = False

        Logger.time = True
        self.total_time = 0.0

        basename = None
        if self.config.smt2file is not None:
            basename = ".".join(self.config.smt2file.split(".")[:-1])
        self.solver = TraceSolver(config.solver_name, "main", basename)

        self.varmapf_t = None
        self.varmapb_t = None

    def unroll(self, trans, invar, k_end, k_start=0, gen_list=False):
        Logger.log("Unroll from %s to %s"%(k_start, k_end), 2)

        fwd = k_start <= k_end
        time_function = self.at_time if fwd else self.at_ptime
        (k_start, k_end) = (min(k_start, k_end), max(k_start, k_end))

        formula = []
        t = k_start
        while t < k_end:
            to_t = t+1 if fwd else t
            formula.append(time_function(trans, t))
            formula.append(time_function(invar, to_t))
            Logger.log("Add trans, k=%s"%t, 2)
            t += 1

        if gen_list:
            return formula
            
        return And(formula)
        
    def _remap_model(self, vars, model, k):
        if model is None:
            return model

        if self.config.strategy == VerificationStrategy.BWD:
            return self._remap_model_bwd(vars, model, k)

        if self.config.strategy == VerificationStrategy.ZZ:
            return self._remap_model_zz(vars, model, k)

        if self.config.strategy in [VerificationStrategy.AUTO, \
                                    VerificationStrategy.FWD, \
                                    VerificationStrategy.NU, \
                                    VerificationStrategy.INT, \
                                    VerificationStrategy.LTL,
                                    VerificationStrategy.ALL]:
            return self._remap_model_fwd(vars, model, k)

        Logger.error("Invalid configuration strategy")
        return None
        
    def _init_at_time(self, vars, maxtime):

        previous = self.config.strategy != VerificationStrategy.FWD

        if self.varmapf_t is not None:
            del(self.varmapf_t)

        if self.varmapb_t is not None:
            del(self.varmapb_t)
            
        self.varmapf_t = {}
        self.varmapb_t = {}

        timed = TS.get_timed_name
        ptimed = TS.get_ptimed_name
        prime = TS.get_prime_name
        prev = TS.get_prev_name

        varsstr = [v.symbol_name() for v in vars]

        for t in range(maxtime+2):
            varmapf = []
            varmapb = []

            for sname in varsstr:
                psname = prime(sname)
                rsname = prev(sname)

                varmapf.append((sname, timed(sname, t)))
                varmapf.append((psname, timed(sname, t+1)))
                varmapf.append((rsname, timed(sname, t-1)))

                if previous:
                    varmapb.append((sname, ptimed(sname, t)))
                    varmapb.append((psname, ptimed(sname, t-1)))
                    varmapb.append((rsname, ptimed(sname, t+1)))

            self.varmapf_t[t] = dict(varmapf)

            if previous:
                self.varmapb_t[t-1] = dict(varmapb)

    def at_time(self, formula, t):
        return substitute(formula, self.varmapf_t[t])

    def at_ptime(self, formula, t):
        return substitute(formula, self.varmapb_t[t])
    
    def _write_smt2_log(self, solver, line):
        tracefile = solver.trace_file
        if tracefile is not None:
            with open(tracefile, "a") as f:
                f.write(line+"\n")

    def _write_smt2_comment(self, solver, line):
        return self._write_smt2_log(solver, ";; %s"%line)

    def _add_assertion(self, solver, formula, comment=None):
        if not self.config.skip_solving:
            solver.solver.add_assertion(formula)

        if Logger.level(3):
            buf = cStringIO()
            printer = SmtPrinter(buf)
            printer.printer(formula)
            print(buf.getvalue()+"\n")

        if solver.trace_file is not None:
            if comment:
                self._write_smt2_comment(solver, "%s: START"%comment)

            formula_fv = get_free_variables(formula)
                
            for v in formula_fv:
                if v in solver.smt2vars:
                    continue
                
                if v.symbol_type() == BOOL:
                    self._write_smt2_log(solver, "(declare-fun %s () Bool)" % (v.symbol_name()))
                elif v.symbol_type().is_array_type():
                    st = v.symbol_type()
                    assert st.index_type.is_bv_type(), "Expecting BV indices"
                    assert st.elem_type.is_bv_type(), "Expecting BV elements"
                    self._write_smt2_log(solver, "(declare-fun %s () (Array (_ BitVec %s) (_ BitVec %s)))"%(v.symbol_name(), st.index_type.width, st.elem_type.width))
                elif v.symbol_type().is_bv_type():
                    self._write_smt2_log(solver, "(declare-fun %s () (_ BitVec %s))" % (v.symbol_name(), v.symbol_type().width))
                else:
                    Logger.error("Unhandled type in smt2 translation")

            self._write_smt2_log(solver, "")

            for v in formula_fv:
                solver.smt2vars.add(v)

            if formula.is_and():
                for f in conjunctive_partition(formula):
                    buf = cStringIO()
                    printer = SmtPrinter(buf)
                    printer.printer(f)
                    self._write_smt2_log(solver, "(assert %s)"%buf.getvalue())
            else:
                buf = cStringIO()
                printer = SmtPrinter(buf)
                printer.printer(formula)
                self._write_smt2_log(solver, "(assert %s)"%buf.getvalue())

            if comment:
                self._write_smt2_comment(solver, "%s: END"%comment)
                                

    def _push(self, solver):
        if not self.config.skip_solving:
            solver.solver.push()

        solver.smt2vars_inc.append(solver.smt2vars)
        self._write_smt2_log(solver, "(push 1)")

    def _pop(self, solver):
        if not self.config.skip_solving:
            solver.solver.pop()

        solver.smt2vars = solver.smt2vars_inc.pop()
        self._write_smt2_log(solver, "(pop 1)")

    def _get_model(self, solver, relevant_vars=None):
        if relevant_vars is None:
            return dict(solver.solver.get_model())

        return dict([(v, solver.solver.get_value(v)) for v in relevant_vars])
        
    def _reset_assertions(self, solver, clear=False):
        if clear:
            solver.clear()
        if not self.config.skip_solving:
            solver.solver.reset_assertions()

        if solver.trace_file is not None:
            solver.smt2vars = set([])
            with open(solver.trace_file, "w") as f:
                f.write("(set-logic %s)\n"%self.hts.logic)

    def _solve(self, solver):
        self._write_smt2_log(solver, "(check-sat)")
        self._write_smt2_log(solver, "")

        if self.config.skip_solving:
            return None

        if Logger.level(2):
            timer = Logger.start_timer("Solve")

        r = solver.solver.solve()

        if Logger.level(2):
            self.total_time += Logger.get_timer(timer)
            Logger.log("Total time solve: %.2f sec"%self.total_time, 1)

        return r
                

    def _check_lemma(self, hts, lemma, init, trans):

        def check_init():
            self._reset_assertions(self.solver)
            self._add_assertion(self.solver, self.at_time(And(init, Not(lemma)), 0), comment="Init check")
            res = self._solve(self.solver)

            prefix = None
            if self.config.prefix is not None:
                prefix = self.config.prefix+"-ind"

            if res:
                Logger.log("Lemma \"%s\" failed for I -> L"%lemma, 2)
                return False
            
            Logger.log("Lemma \"%s\" holds for I -> L"%lemma, 2)
            return True

    
    def _suff_lemmas(self, prop, lemmas):
        self._reset_assertions(self.solver)

        self._add_assertion(self.solver, And(And(lemmas), Not(prop)))
        
        if self._solve(self.solver):
            return False

        return True

    def add_lemmas(self, hts, prop, lemmas):
        if len(lemmas) == 0:
            return (hts, False)

        self._reset_assertions(self.solver)

        h_init = hts.single_init()
        h_trans = hts.single_trans()
        
        holding_lemmas = []
        lindex = 1
        nlemmas = len(lemmas)
        tlemmas = 0
        flemmas = 0
        for lemma in lemmas:
            Logger.log("\nChecking Lemma %s/%s"%(lindex,nlemmas), 1)
            invar = hts.single_invar()
            init = And(h_init, invar)
            trans = And(invar, h_trans, TS.to_next(invar))
            if self._check_lemma(hts, lemma, init, trans):
                holding_lemmas.append(lemma)
                hts.add_assumption(lemma)
                hts.reset_formulae()
                
                Logger.log("Lemma %s holds"%(lindex), 1)
                tlemmas += 1
                if self._suff_lemmas(prop, holding_lemmas):
                    return (hts, True)
            else:
                Logger.log("Lemma %s does not hold"%(lindex), 1)
                flemmas += 1
                
            msg = "%s T:%s F:%s U:%s"%(status_bar((float(lindex)/float(nlemmas)), False), tlemmas, flemmas, (nlemmas-lindex))
            Logger.inline(msg, 0, not(Logger.level(1))) 
            lindex += 1
            
        Logger.clear_inline(0, not(Logger.level(1)))
        
        hts.assumptions = And(holding_lemmas)
        return (hts, False)
    
    def _remap_model_fwd(self, vars, model, k):
        return model

    def _remap_model_bwd(self, vars, model, k):
        retmodel = dict()

        for var in vars:
            for t in range(k+1):
                retmodel[TS.get_timed(var, t)] = model[TS.get_ptimed(var, k-t)]

        return retmodel

    def _remap_model_zz(self, vars, model, k):
        retmodel = dict([el for el in dict(model).items() if not TS.is_ptimed(el[0])])

        for var in vars:
            for t in range(int(k/2)+1, k+1, 1):
                retmodel[TS.get_timed(var, t)] = model[TS.get_ptimed(var, k-t)]

        return retmodel

    def generate_trace(self, \
                       model, \
                       length, \
                       xvars=None, \
                       find_loop=False):

        trace = Trace()
        trace.model = model
        trace.length = length
        trace.infinite = find_loop
        trace.prop_vars = xvars
        
        return trace
    



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pysmt.shortcuts import Not, TRUE, And, BVNot, BVAnd, BVOr, BVAdd, Or, Symbol, BV, EqualsOrIff, \
    Implies, BVMul, BVExtract, BVUGT, BVULT, BVULE, Ite, BVZExt, BVXor, BVConcat, get_type, BVSub, \
    Xor, Select, Store
from pysmt.typing import BOOL, BVType, ArrayType

from cosa.representation import HTS, TS
from cosa.encoders.formulae import StringParser
from cosa.utils.logger import Logger
from cosa.utils.formula_mngm import quote_names, B2BV, BV2B
from cosa.utils.generic import bin_to_dec
from cosa.encoders.template import ModelParser

NL = "\n"

SN="N%s"

COM=";"
SORT="sort"
BITVEC="bitvec"
ARRAY="array"
WRITE="write"
READ="read"
ZERO="zero"
ONE="one"
ONES="ones"
STATE="state"
INPUT="input"
OUTPUT="output"
ADD="add"
EQ="eq"
NE="ne"
MUL="mul"
SLICE="slice"
CONST="const"
CONSTD="constd"
UGT="ugt"
ULT="ult"
ULTE="ulte"
AND="and"
XOR="xor"
NAND="nand"
IMPLIES="implies"
OR="or"
ITE="ite"
NOT="not"
REDOR="redor"
REDAND="redand"
UEXT="uext"
CONCAT="concat"
SUB="sub"

INIT="init"
NEXT="next"
CONSTRAINT="constraint"
BAD="bad"

class BTOR2Parser(ModelParser):
    parser = None
    extensions = ["btor2","btor"]
    name = "BTOR2"
    
    def __init__(self):
        pass

    def parse_file(self, strfile, config, flags=None):
        with open(strfile, "r") as f:
            return self.parse_string(f.read())

    def is_available(self):
        return True
        
    def get_extensions(self):
        return self.extensions

    @staticmethod
    def get_extensions():
        return BTOR2Parser.extensions

    def remap_an2or(self, name):
        return name

    def remap_or2an(self, name):
        return name

    def parse_string(self, strinput):

        hts = HTS()
        ts = TS()

        nodemap = {}
        node_covered = set([])

        translist = []
        initlist = []
        invarlist = []

        invar_props = []
        ltl_props = []

        def getnode(nid):
            node_covered.add(nid)
            if int(nid) < 0:
                return Ite(BV2B(nodemap[str(-int(nid))]), BV(0,1), BV(1,1))
            return nodemap[nid]

        def binary_op(bvop, bop, left, right):
            if (get_type(left) == BOOL) and (get_type(right) == BOOL):
                return bop(left, right)
            return bvop(B2BV(left), B2BV(right))

        def unary_op(bvop, bop, left):
            if (get_type(left) == BOOL):
                return bop(left)
            return bvop(left)

        for line in strinput.split(NL):
            linetok = line.split()
            if len(linetok) == 0:
                continue
            if linetok[0] == COM:
                continue

            (nid, ntype, *nids) = linetok

            if ntype == SORT:
                (stype, *attr) = nids
                if stype == BITVEC:
                    nodemap[nid] = BVType(int(attr[0]))
                    node_covered.add(nid)
                if stype == ARRAY:
                    nodemap[nid] = ArrayType(getnode(attr[0]), getnode(attr[1]))
                    node_covered.add(nid)

            if ntype == WRITE:
                nodemap[nid] = Store(*[getnode(n) for n in nids[1:4]])

            if ntype == READ:
                nodemap[nid] = Select(getnode(nids[1]), getnode(nids[2]))

            if ntype == ZERO:
                nodemap[nid] = BV(0, getnode(nids[0]).width)

            if ntype == ONE:
                nodemap[nid] = BV(1, getnode(nids[0]).width)

            if ntype == ONES:
                width = getnode(nids[0]).width
                nodemap[nid] = BV((2**width)-1, width)

            if ntype == REDOR:
                width = get_type(getnode(nids[1])).width
                zeros = BV(0, width)
                nodemap[nid] = Not(EqualsOrIff(getnode(nids[1]), zeros))

            if ntype == REDAND:
                width = get_type(getnode(nids[1])).width
                ones = BV((2**width)-1, width)
                nodemap[nid] = EqualsOrIff(getnode(nids[1]), ones)

            if ntype == CONSTD:
                width = getnode(nids[0]).width
                nodemap[nid] = BV(int(nids[1]), width)

            if ntype == CONST:
                width = getnode(nids[0]).width
                nodemap[nid] = BV(bin_to_dec(nids[1]), width)

            if ntype == STATE:
                if len(nids) > 1:
                    nodemap[nid] = Symbol(nids[1], getnode(nids[0]))
                else:
                    nodemap[nid] = Symbol((SN%nid), getnode(nids[0]))
                ts.add_state_var(nodemap[nid])

            if ntype == INPUT:
                if len(nids) > 1:
                    nodemap[nid] = Symbol(nids[1], getnode(nids[0]))
                else:
                    nodemap[nid] = Symbol((SN%nid), getnode(nids[0]))
                ts.add_input_var(nodemap[nid])

            if ntype == OUTPUT:
                if len(nids) > 2:
                    symbol = Symbol(nids[2], getnode(nids[0]))
                else:
                    symbol = Symbol((SN%nid), getnode(nids[0]))

                nodemap[nid] = EqualsOrIff(symbol, B2BV(getnode(nids[1])))
                invarlist.append(nodemap[nid])
                node_covered.add(nid)
                ts.add_output_var(symbol)

            if ntype == AND:
                nodemap[nid] = binary_op(BVAnd, And, getnode(nids[1]), getnode(nids[2]))

            if ntype == CONCAT:
                nodemap[nid] = BVConcat(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == XOR:
                nodemap[nid] = binary_op(BVXor, Xor, getnode(nids[1]), getnode(nids[2]))

            if ntype == NAND:
                bvop = lambda x,y: BVNot(BVAnd(x, y))
                bop = lambda x,y: Not(And(x, y))
                nodemap[nid] = binary_op(bvop, bop, getnode(nids[1]), getnode(nids[2]))

            if ntype == IMPLIES:
                nodemap[nid] = Implies(BV2B(getnode(nids[1])), BV2B(getnode(nids[2])))

            if ntype == NOT:
                nodemap[nid] = unary_op(BVNot, Not, getnode(nids[1]))

            if ntype == UEXT:
                nodemap[nid] = BVZExt(B2BV(getnode(nids[1])), int(nids[2]))

            if ntype == OR:
                nodemap[nid] = binary_op(BVOr, Or, getnode(nids[1]), getnode(nids[2]))

            if ntype == ADD:
                nodemap[nid] = BVAdd(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == SUB:
                nodemap[nid] = BVSub(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == UGT:
                nodemap[nid] = BVUGT(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == ULT:
                nodemap[nid] = BVULT(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == ULTE:
                nodemap[nid] = BVULE(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == EQ:
                nodemap[nid] = EqualsOrIff(getnode(nids[1]), getnode(nids[2]))

            if ntype == NE:
                nodemap[nid] = Not(EqualsOrIff(getnode(nids[1]), getnode(nids[2])))

            if ntype == MUL:
                nodemap[nid] = BVMul(B2BV(getnode(nids[1])), B2BV(getnode(nids[2])))

            if ntype == SLICE:
                nodemap[nid] = BVExtract(B2BV(getnode(nids[1])), int(nids[3]), int(nids[2]))

            if ntype == ITE:
                if (get_type(getnode(nids[2])) == BOOL) or (get_type(getnode(nids[3])) == BOOL):
                    nodemap[nid] = Ite(BV2B(getnode(nids[1])), BV2B(getnode(nids[2])), BV2B(getnode(nids[3])))
                else:
                    nodemap[nid] = Ite(BV2B(getnode(nids[1])), getnode(nids[2]), getnode(nids[3]))

            if ntype == NEXT:
                if (get_type(getnode(nids[1])) == BOOL) or (get_type(getnode(nids[2])) == BOOL):
                    nodemap[nid] = EqualsOrIff(BV2B(TS.get_prime(getnode(nids[1]))), BV2B(getnode(nids[2])))
                else:
                    nodemap[nid] = EqualsOrIff(TS.get_prime(getnode(nids[1])), getnode(nids[2]))
                translist.append(getnode(nid))

            if ntype == INIT:
                if (get_type(getnode(nids[1])) == BOOL) or (get_type(getnode(nids[2])) == BOOL):
                    nodemap[nid] = EqualsOrIff(BV2B(getnode(nids[1])), BV2B(getnode(nids[2])))
                else:
                    nodemap[nid] = EqualsOrIff(getnode(nids[1]), getnode(nids[2]))
                initlist.append(getnode(nid))

            if ntype == CONSTRAINT:
                nodemap[nid] = BV2B(getnode(nids[0]))
                invarlist.append(getnode(nid))

            if ntype == BAD:
                nodemap[nid] = getnode(nids[0])
                invar_props.append(Not(BV2B(getnode(nid))))

            if nid not in nodemap:
                Logger.error("Unknown node type \"%s\""%ntype)

        if Logger.level(1):
            name = lambda x: str(nodemap[x]) if nodemap[x].is_symbol() else x
            uncovered = [name(x) for x in nodemap if x not in node_covered]
            uncovered.sort()
            if len(uncovered) > 0:
                Logger.warning("Unlinked nodes \"%s\""%",".join(uncovered))
        
        init = And(initlist)
        trans = And(translist)
        invar = And(invarlist)

        ts.set_behavior(init, trans, invar)
        hts.add_ts(ts)

        return (hts, invar_props, ltl_props)



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import re

from cosa.representation import HTS, TS
from cosa.utils.logger import Logger

from cosa.encoders.template import ModelParser
from cosa.encoders.verilog_hts import VerilogHTSParser

from cosa.utils.generic import suppress_output, restore_output

COMMANDS = []
COMMANDS.append("read -sysv {FILES} -top {TARGET}")
COMMANDS.append("write -format verilog {FILE}")

EXPIRED = "asdads"

TMPFILE = "__verific_verilog__.v"
TMPCMDFILE = "__verific_commands__.txt"

CMD = "verific"

class SystemVerilogVerificParser(ModelParser):
    parser = None
    extensions = ["sv"]
    name = "SystemVerilog (Verific)"

    model_info = None

    def __init__(self):
        pass

    def is_available(self):
        if shutil.which(CMD) is None:
            return False

        print_level = 3
        if not Logger.level(print_level):
            saved_stdout = suppress_output()
        
        retval = os.system("%s -h"%CMD)

        if not Logger.level(print_level):
            restore_output(saved_stdout)

        if retval != 0:
            return False
        
        return True

    def get_model_info(self):
        return self.model_info
     
    def parse_file(self, strfile, config, flags=None):
        if flags is None:
            Logger.error("Top module not provided")

        topmodule = flags[0]
        absstrfile = os.path.abspath(strfile)
        directory = "/".join(absstrfile.split("/")[:-1])
        filename = absstrfile.split("/")[-1]

        files = [absstrfile]

        commands = "\n".join(COMMANDS)
        commands = commands.format(FILES=" ".join(files), \
                                   TARGET=topmodule, \
                                   FILE=TMPFILE)

        command = "%s -script_file %s"%(CMD, TMPCMDFILE)
        
        with open(TMPCMDFILE, "w") as f:
            f.write(commands)
        
        Logger.log("Commands: %s"%commands, 2)

        print_level = 3
        if not Logger.level(print_level):
            saved_stdout = suppress_output()
        
        retval = os.system(command)

        if not Logger.level(print_level):
            restore_output(saved_stdout)

        if retval != 0:
            Logger.error("Error in SystemVerilog conversion")
            
        parser = VerilogHTSParser()
        ret = parser.parse_file(TMPFILE, config, flags=flags)
        self.model_info = parser.get_model_info()

        if (not Logger.level(1)) and (not config.debug):
            os.remove(TMPFILE)
            os.remove(TMPCMDFILE)
        
        return ret

    def get_extensions(self):
        return self.extensions

    @staticmethod        
    def get_extensions():
        return SystemVerilogVerificParser.extensions

    def remap_an2or(self, name):
        return name

    def remap_or2an(self, name):
        return name
    
    def parse_string(self, strinput):
        return

        return (hts, invar_props, ltl_props)



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from pysmt.shortcuts import And, Or, TRUE, FALSE, Not, EqualsOrIff, Implies, Iff, Symbol, BOOL, \
    BV, BVAdd, BVSub, get_type, BVULT, BVUGT, BVULE
from pysmt.typing import BOOL, BVType, ArrayType
from pysmt.fnode import FNode

from cosa.encoders.template import STSGenerator
from cosa.representation import TS
from cosa.utils.logger import Logger
from cosa.encoders.formulae import StringParser
from cosa.utils.formula_mngm import BV2B

class FixedScoreBoardGenerator(STSGenerator):
    name = "FixedScoreboard"
    description = "Scoreboard for a fifo with no pop"
    interface = "input_port, max_value, clock"

    def get_param_length(self):
        return 3
    
    def compile_sts(self, name, params):
        in_port, max_val, clk = list(params)

        sb = ScoreBoardGenerator()
        return sb.compile_sts(name, [in_port, max_val, clk, FALSE()])

class ScoreBoardGenerator(STSGenerator):
    name = "Scoreboard"
    description = "Scoreboard for a fifo with push and pop"
    interface = "input_port, max_value, signal_push, signal_pop"

    def get_param_length(self):
        return 4

    def compile_sts(self, name, params):
        sparser = StringParser()
        in_port, max_val, c_push, c_pop = list(params)
        max_val = int(max_val)

        if type(c_push) == str:
            c_push = sparser.parse_formula(c_push)
        if type(c_pop) == str:
            c_pop = sparser.parse_formula(c_pop)

        tracking = Symbol("%s.tracking"%name, BOOL)
        end = Symbol("%s.end"%name, BOOL)
        done = Symbol("%s.done"%name, BOOL)
        packet = Symbol("%s.packet"%name, BVType(in_port.symbol_type().width))
        max_width = math.ceil(math.log(max_val)/math.log(2))

        max_bvval = BV(max_val, max_width)
        zero = BV(0, max_width)
        one = BV(1, max_width)
        count = Symbol("%s.count"%name, BVType(max_width))
        size = Symbol("%s.size"%name, BVType(max_width))
        
        pos_c_push = BV2B(c_push)
        neg_c_push = Not(BV2B(c_push))

        pos_c_pop = BV2B(c_pop)
        neg_c_pop = Not(BV2B(c_pop))

        init = []
        trans = []
        invar = []
        
        # INIT DEFINITION #

        # count = 0
        init.append(EqualsOrIff(count, BV(0, max_width)))
        # tracking = False
        init.append(EqualsOrIff(tracking, FALSE()))
        # size = 0
        init.append(EqualsOrIff(size, BV(0, max_width)))
        # end = false
        init.append(EqualsOrIff(end, FALSE()))

        # INVAR DEFINITION #

        # !done -> (end = (tracking & (size = count)))
        invar.append(Implies(Not(done), EqualsOrIff(end, And(tracking, EqualsOrIff(size, count)))))
        
        # count <= size
        invar.append(BVULE(count, size))
        # count <= maxval
        invar.append(BVULE(count, max_bvval))
        # size <= maxval
        invar.append(BVULE(size, max_bvval))

        # done -> (end <-> False);
        invar.append(Implies(done, EqualsOrIff(end, FALSE())))
        # done -> (count = 0_8);
        invar.append(Implies(done, EqualsOrIff(count, BV(0, max_width))))
        # done -> (size = 0_8);
        invar.append(Implies(done, EqualsOrIff(size, BV(0, max_width))))
        # done -> (packet = 0_8);
        invar.append(Implies(done, EqualsOrIff(packet, BV(0, in_port.symbol_type().width))))
        
        # TRANS DEFINITION #

        # (!end & !done) -> next(!done);
        trans.append(Implies(And(Not(end), Not(done)), TS.to_next(Not(done))))
        # end -> next(done);
        trans.append(Implies(end, TS.to_next(done)))
        # done -> next(done);
        trans.append(Implies(done, TS.to_next(done)))
        
        # tracking -> next(tracking);
        trans.append(Implies(Not(done), Implies(tracking, TS.to_next(tracking))))
        # tracking -> (next(packet) = packet);
        trans.append(Implies(Not(done), Implies(tracking, EqualsOrIff(TS.to_next(packet), packet))))
        # !tracking & next(tracking) -> c_push;
        trans.append(Implies(Not(done), Implies(And(Not(tracking), TS.to_next(tracking)), pos_c_push)))
        # (c_push & next(tracking)) -> ((packet = in) & (next(packet) = in);
        trans.append(Implies(Not(done), Implies(And(pos_c_push, TS.to_next(tracking)), And(EqualsOrIff(packet, in_port), EqualsOrIff(TS.to_next(packet), in_port)))))
        # (c_push & !c_pop & tracking) -> (next(count) = (count + 1_8));
        trans.append(Implies(Not(done), Implies(And(pos_c_push, neg_c_pop, tracking), EqualsOrIff(TS.to_next(count), BVAdd(count, BV(1, max_width))))))
        # (c_push & size < maxval) -> (next(size) = (size + 1_8));
        trans.append(Implies(Not(done), Implies(And(pos_c_push, BVULT(size, max_bvval)), EqualsOrIff(TS.to_next(size), BVAdd(size, BV(1, max_width))))))
        # (c_pop & size > 0) -> (next(size) = (size - 1_8));
        trans.append(Implies(Not(done), Implies(And(pos_c_pop, BVUGT(size, zero)), EqualsOrIff(TS.to_next(size), BVSub(size, BV(1, max_width))))))
        # (!(c_push | c_pop)) -> (next(count) = count);
        trans.append(Implies(Not(done), Implies(Not(Or(pos_c_push, pos_c_pop)), EqualsOrIff(count, TS.to_next(count)))))
        # ((c_push | c_pop) & !tracking) -> (next(count) = count);
        trans.append(Implies(Not(done), Implies(And(Or(pos_c_push, pos_c_pop), Not(tracking)), EqualsOrIff(count, TS.to_next(count)))))

        # (c_push & size = maxval) -> (next(size) = size);
        trans.append(Implies(Not(done), Implies(And(pos_c_push, EqualsOrIff(size, max_bvval)), EqualsOrIff(TS.to_next(size), size))))
        # (!(c_push | c_pop)) -> (next(size) = size);
        trans.append(Implies(Not(done), Implies(Not(Or(pos_c_push, pos_c_pop)), EqualsOrIff(size, TS.to_next(size)))))
        # (!(c_push | c_pop)) -> (next(count) = count);
        trans.append(Implies(Not(done), Implies(Not(Or(pos_c_push, pos_c_pop)), EqualsOrIff(count, TS.to_next(count)))))
        # (c_pop & size = 0) -> (next(size) = 0);
        trans.append(Implies(Not(done), Implies(And(pos_c_pop, EqualsOrIff(size, zero)), EqualsOrIff(TS.to_next(size), zero))))

        # (!c_push) -> (next(count) = count);
        trans.append(Implies(Not(done), Implies(neg_c_push, EqualsOrIff(TS.to_next(count), count))))
        
        init = And(init)
        invar = And(invar)
        trans = And(trans)

        ts = TS()
        ts.vars, ts.init, ts.invar, ts.trans = set([tracking, end, packet, count, size]), init, invar, trans

        return ts
    
class RandomGenerator(STSGenerator):
    name = "Random"
    description = "Random Generator"
    interface = "varsize"

    def get_param_length(self):
        return 1
    
    def compile_sts(self, name, params):
        ts = TS()
        parsize = params[0]
        size = None
        
        if type(parsize) == str:
            sparser = StringParser()
            parsize = sparser.parse_formula(parsize)

        if parsize.is_constant():
            size = parsize.constant_value()

        if get_type(parsize).is_bv_type():
            size = get_type(parsize).width
            
        if size is None:
            Logger.error("Undefined size for symbol \"%s\""%(params[0]))

        value = Symbol("%s.value"%name, BVType(size))
        ts.add_var(value)
        ts.trans = EqualsOrIff(value, TS.get_prime(value))

        return ts
    



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy

from cosa.utils.logger import Logger
from cosa.analyzers.mcsolver import MCConfig
from cosa.analyzers.bmc_safety import BMCSafety
from cosa.analyzers.bmc_ltl import BMCLTL
from cosa.problem import VerificationType, Problem, VerificationStatus, Trace
from cosa.encoders.miter import Miter
from cosa.encoders.formulae import StringParser
from cosa.representation import HTS, TS
from cosa.encoders.explicit_transition_system import ExplicitTSParser
from cosa.encoders.symbolic_transition_system import SymbolicTSParser, SymbolicSimpleTSParser
from cosa.encoders.btor2 import BTOR2Parser
from cosa.encoders.ltl import LTLParser
from cosa.encoders.factory import ModelParsersFactory, ClockBehaviorsFactory, GeneratorsFactory
from cosa.encoders.template import EncoderConfig, ModelInformation
from cosa.encoders.parametric_behavior import ParametricBehavior
from cosa.printers.trace import TextTracePrinter, VCDTracePrinter

FLAG_SR = "["
FLAG_ST = "]"
FLAG_SP = "+"

MODEL_SP = ";"
FILE_SP  = ","

class ProblemSolver(object):
    parser = None
    sparser = None
    lparser = None
    model_info = None

    def __init__(self):
        self.sparser = None
        self.lparser = None
        self.model_info = ModelInformation()

        GeneratorsFactory.init_generators()
        ClockBehaviorsFactory.init_clockbehaviors()

    def __process_trace(self, hts, trace, config, problem):
        prevass = []

        full_trace = problem.full_trace or config.full_trace
        trace_vars_change = problem.trace_vars_change or config.trace_vars_change
        trace_all_vars = problem.trace_all_vars or config.trace_all_vars

        diff_only = not trace_vars_change
        all_vars = trace_all_vars
        
        txttrace_synth_clock = False

        if full_trace:
            diff_only = False
            all_vars = True

        traces = []
        abstract_clock_list = []

        if txttrace_synth_clock:
            abstract_clock_list=self.model_info.abstract_clock_list
        
        # Human Readable Format
        hr_printer = TextTracePrinter()
        hr_printer.prop_vars = trace.prop_vars
        hr_printer.diff_only = diff_only
        hr_printer.all_vars = all_vars
        hr_trace = hr_printer.print_trace(hts=hts, \
                                          model=trace.model, \
                                          length=trace.length, \
                                          map_function=self.parser.remap_an2or, \
                                          find_loop=trace.infinite, \
                                          abstract_clock_list=abstract_clock_list)
        traceH = Trace(hr_trace, trace.length)
        traceH.extension = hr_printer.get_file_ext()
        traceH.human_readable = hr_trace.human_readable
        traces.append(traceH)

        # VCD format
        vcd_trace = None
        if problem.vcd:
            vcd_printer = VCDTracePrinter()
            vcd_trace = vcd_printer.print_trace(hts=hts, \
                                                model=trace.model, \
                                                length=trace.length, \
                                                map_function=self.parser.remap_an2or, \
                                                abstract_clock_list=self.model_info.abstract_clock_list)
            traceV = Trace(vcd_trace, trace.length)
            traceV.extension = vcd_printer.get_file_ext()
            traceV.human_readable = vcd_trace.human_readable
            traces.append(traceV)

        return traces
        
    def __solve_problem(self, problem, config):
        if problem.name is not None:
            Logger.log("\n*** Analyzing problem \"%s\" ***"%(problem), 1)
            Logger.msg("Solving \"%s\" "%problem.name, 0, not(Logger.level(1)))
        
        mc_config = self.problem2mc_config(problem, config)
        bmc_safety = BMCSafety(problem.hts, mc_config)
        bmc_ltl = BMCLTL(problem.hts, mc_config)
        res = VerificationStatus.UNC
        bmc_length = max(problem.bmc_length, config.bmc_length)
        bmc_length_min = max(problem.bmc_length_min, config.bmc_length_min)

        parsing_defs = [problem.formula, problem.lemmas, problem.assumptions]
        for i in range(len(parsing_defs)):
            if parsing_defs[i] is not None:
                pdef_file = problem.relative_path+parsing_defs[i]
                if os.path.isfile(pdef_file):
                    with open(pdef_file) as f:
                        parsing_defs[i] = [p.strip() for p in f.read().strip().split("\n")]
                else:
                    parsing_defs[i] = [p.strip() for p in parsing_defs[i].split(MODEL_SP)]
            else:
                parsing_defs[i] = None

        [formulae, problem.lemmas, problem.assumptions] = parsing_defs

        ParametricBehavior.apply_to_problem(problem, self.model_info)

        assumps = None
        lemmas = None

        if formulae is None:
            if problem.verification == VerificationType.SIMULATION:
                formulae = ["True"]
            elif (problem.verification is not None) and (problem.verification != VerificationType.EQUIVALENCE):
                Logger.error("Property not provided")
                
        accepted_ver = False

        if formulae is not None:
            problem.formula = formulae[0]
        
        precondition = config.precondition if config.precondition is not None else problem.precondition
        
        if precondition and problem.verification == VerificationType.SAFETY:
            problem.formula = "(%s) -> (%s)"%(precondition, problem.formula)
        
        if (problem.verification != VerificationType.EQUIVALENCE) and (problem.formula is not None):
            assumps = [t[1] for t in self.sparser.parse_formulae(problem.assumptions)]
            lemmas = [t[1] for t in self.sparser.parse_formulae(problem.lemmas)]
            for ass in assumps:
                problem.hts.add_assumption(ass)
            for lemma in lemmas:
                problem.hts.add_lemma(lemma)
            if problem.verification != VerificationType.LTL:
                (strprop, prop, types) = self.sparser.parse_formulae([problem.formula])[0]
            else:
                (strprop, prop, types) = self.lparser.parse_formulae([problem.formula])[0]

            problem.formula = prop

        if problem.verification is None:
            return problem
            
        if problem.verification == VerificationType.SAFETY:
            accepted_ver = True
            Logger.log("Property: %s"%(prop.serialize(threshold=100)), 2)
            res, trace, _ = bmc_safety.safety(prop, bmc_length, bmc_length_min)

        if problem.verification == VerificationType.LTL:
            accepted_ver = True
            res, trace, _ = bmc_ltl.ltl(prop, bmc_length, bmc_length_min)

        if problem.verification == VerificationType.SIMULATION:
            accepted_ver = True
            res, trace = bmc_safety.simulate(prop, bmc_length)

        hts = problem.hts
            
        if problem.verification == VerificationType.EQUIVALENCE:
            accepted_ver = True
            if problem.equivalence:
                (problem.hts2, _, _) = self.parse_model(problem.relative_path, \
                                                        problem.equivalence, \
                                                        encoder_config, \
                                                        "System 2")

            htseq, miter_out = Miter.combine_systems(problem.hts, \
                                                     problem.hts2, \
                                                     bmc_length, \
                                                     problem.symbolic_init, \
                                                     problem.formula, \
                                                     True)

            if problem.assumptions is not None:
                assumps = [t[1] for t in self.sparser.parse_formulae(problem.assumptions)]

            if problem.lemmas is not None:
                lemmas = [t[1] for t in self.sparser.parse_formulae(problem.lemmas)]

            if assumps is not None:
                for assumption in assumps:
                    htseq.add_assumption(assumption)

            if lemmas is not None:
                for lemma in lemmas:
                    htseq.add_lemma(lemma)

            bmcseq = BMCSafety(htseq, mc_config)
            hts = htseq
            res, trace, t = bmcseq.safety(miter_out, bmc_length, bmc_length_min)

        if not accepted_ver:
            Logger.error("Invalid verification type")
            
        problem.status = res
        if trace is not None:
            problem.traces = self.__process_trace(hts, trace, config, problem)

        if problem.assumptions is not None:
            problem.hts.assumptions = None

        Logger.log("\n*** Problem \"%s\" is %s ***"%(problem, res), 1)

    def get_file_flags(self, strfile):
        if FLAG_SR not in strfile:
            return (strfile, None)
        
        (strfile, flags) = (strfile[:strfile.index(FLAG_SR)], strfile[strfile.index(FLAG_SR)+1:strfile.index(FLAG_ST)].split(FLAG_SP))
        return (strfile, flags)
        
    def parse_model(self, \
                    relative_path, \
                    model_files, \
                    encoder_config, \
                    name=None):
        
        hts = HTS("System 1")
        invar_props = []
        ltl_props = []

        models = model_files.split(FILE_SP)

        for strfile in models:
            (strfile, flags) = self.get_file_flags(strfile)
            filetype = strfile.split(".")[-1]
            strfile = strfile.replace("~", os.path.expanduser("~"))
            if strfile[0] != "/":
                strfile = relative_path+strfile
            parser = None

            for av_parser in ModelParsersFactory.get_parsers():
                assert av_parser.name is not None
                if filetype in av_parser.get_extensions():
                    parser = av_parser
                    if not self.parser:
                        self.parser = av_parser
                    
            if parser is not None:
                if not os.path.isfile(strfile):
                    Logger.error("File \"%s\" does not exist"%strfile)

                Logger.msg("Parsing file \"%s\"... "%(strfile), 0)
                (hts_a, inv_a, ltl_a) = parser.parse_file(strfile, encoder_config, flags)
                self.model_info.combine(parser.get_model_info())
                hts.combine(hts_a)

                invar_props += inv_a
                ltl_props += ltl_a
                
                Logger.log("DONE", 0)
                continue

            Logger.error("Filetype \"%s\" unsupported or parser is not available"%filetype)

        if Logger.level(1):
            print(hts.print_statistics(name, Logger.level(2)))

        return (hts, invar_props, ltl_props)

    def solve_problems(self, problems, config):
        encoder_config = self.problems2encoder_config(config, problems)

        self.sparser = StringParser(encoder_config)
        self.lparser = LTLParser()

        invar_props = []
        ltl_props = []
        si = False

        if len(problems.symbolic_inits) == 0:
            problems.symbolic_inits.add(si)

        HTSM = 0
        HTS2 = 1
        HTSD = (HTSM, si)
            
        # generate systems for each problem configuration
        systems = {}
        for si in problems.symbolic_inits:
            encoder_config.symbolic_init = si
            (systems[(HTSM, si)], invar_props, ltl_props) = self.parse_model(problems.relative_path, \
                                                                              problems.model_file, \
                                                                              encoder_config, \
                                                                              "System 1")
            
        if problems.equivalence is not None:
            (systems[(HTS2, si)], _, _) = self.parse_model(problems.relative_path, \
                                                             problems.equivalence, \
                                                             encoder_config, \
                                                             "System 2")
        else:
            systems[(HTS2, si)] = None

        if config.safety or config.problems:
            for invar_prop in invar_props:
                inv_prob = problems.new_problem()
                inv_prob.verification = VerificationType.SAFETY
                inv_prob.name = invar_prop[0]
                inv_prob.description = invar_prop[1]
                inv_prob.formula = invar_prop[2]
                problems.add_problem(inv_prob)

        if config.ltl or config.problems:
            for ltl_prop in ltl_props:
                ltl_prob = problems.new_problem()
                ltl_prob.verification = VerificationType.LTL
                ltl_prob.name = ltl_prop[0]
                ltl_prob.description = ltl_prop[1]
                ltl_prob.formula = ltl_prop[2]
                problems.add_problem(ltl_prob)
                
        assume_if_true = config.assume_if_true or problems.assume_if_true

        if HTSD in systems:
            problems._hts = systems[HTSD]
        
        for problem in problems.problems:
            problem.hts = systems[(HTSM, problem.symbolic_init)]
            if problems._hts is None:
                problems._hts = problem.hts
            problem.hts2 = systems[(HTS2, problem.symbolic_init)]
            if problems._hts2 is None:
                problems._hts2 = problem.hts2
            problem.abstract_clock = problems.abstract_clock or config.abstract_clock
            problem.add_clock = problems.add_clock or config.add_clock
            problem.run_coreir_passes = problems.run_coreir_passes
            problem.relative_path = problems.relative_path

            if not problem.full_trace:
                problem.full_trace = problems.full_trace
            if not problem.trace_vars_change:
                problem.trace_vars_change = problems.trace_vars_change
            if not problem.trace_all_vars:
                problem.trace_all_vars = problems.trace_all_vars
            if not problem.clock_behaviors:
                clk_bhvs = [p for p in [problems.clock_behaviors, config.clock_behaviors] if p is not None]
                if len(clk_bhvs) > 0:
                    problem.clock_behaviors = ";".join(clk_bhvs)
            if not problem.generators:
                problem.generators = config.generators
                
            Logger.log("Solving with abstract_clock=%s, add_clock=%s"%(problem.abstract_clock, problem.add_clock), 2)
            
            if problem.trace_prefix is not None:
                problem.trace_prefix = "".join([problem.relative_path,problem.trace_prefix])
                
            if config.time or problems.time:
                timer_solve = Logger.start_timer("Problem %s"%problem.name, False)
            try:
                self.__solve_problem(problem, config)
                
                if problem.verification is None:
                    Logger.log("Unset verification", 2)
                    continue
                
                Logger.msg(" %s\n"%problem.status, 0, not(Logger.level(1)))

                if (assume_if_true) and \
                   (problem.status == VerificationStatus.TRUE) and \
                   (problem.assumptions == None) and \
                   (problem.verification == VerificationType.SAFETY):

                    ass_ts = TS("Previous assumption from property")
                    if TS.has_next(problem.formula):
                        ass_ts.trans = problem.formula
                    else:
                        ass_ts.invar = problem.formula
                    problem.hts.reset_formulae()

                    problem.hts.add_ts(ass_ts)
                        
                if config.time or problems.time:
                    problem.time = Logger.get_timer(timer_solve, False)
                
            except KeyboardInterrupt as e:
                Logger.msg("\b\b Skipped!\n", 0)

    def problem2mc_config(self, problem, config):
        mc_config = MCConfig()

        config_selection = lambda problem, config: config if problem is None else problem
        
        mc_config.smt2file = config_selection(problem.smt2_tracing, config.smt2file)
        mc_config.prefix = problem.name
        mc_config.strategy = config_selection(problem.strategy, config.strategy)
        mc_config.incremental = config_selection(problem.incremental, config.incremental)
        mc_config.skip_solving = config_selection(problem.skip_solving, config.skip_solving)
        mc_config.solver_name = config_selection(problem.solver_name, config.solver_name)
        mc_config.prove = config_selection(problem.prove, config.prove)

        return mc_config


    def problems2encoder_config(self, config, problems):
        encoder_config = EncoderConfig()
        encoder_config.abstract_clock = problems.abstract_clock or config.abstract_clock
        encoder_config.symbolic_init = config.symbolic_init or config.symbolic_init
        encoder_config.zero_init = problems.zero_init or config.zero_init
        encoder_config.add_clock = problems.add_clock or config.add_clock
        encoder_config.deterministic = config.deterministic
        encoder_config.run_passes = config.run_passes
        encoder_config.boolean = problems.boolean or config.boolean
        encoder_config.debug = config.debug

        return encoder_config
        



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
import sys

STRING_PATTERN = "___STRING_%d___"
string_id = 0

def is_number(strnum):
    try:
        int(strnum)
        return True
    except:
        return False

def auto_convert(strval):
    if (strval.upper() == "TRUE") or (strval.upper() == "YES"):
        return True
    if (strval.upper() == "FALSE") or (strval.upper() == "NO"):
        return False
    try:
        return int(strval)
    except Exception:
        try:
            return float(strval)
        except Exception:
            return strval
    
def status_bar(status, percent=True, length=40):
    curr = math.ceil(length*status)
    percent = (" %.2f%%"%(status*100)) if percent else ""
    return "["+("#"*(curr))+(" "*(length-curr))+"]"+percent

def dec_to_bin(val, width):
    bitval = "{0:b}".format(int(val))
    bitval = "%s%s"%("0"*(width-len(bitval)), bitval)
    return bitval

def dec_to_hex(val, width):
    hexval = str(hex(val))[2:]
    hexval = "%s%s"%("0"*(width-len(hexval)), hexval)
    return hexval.upper()

def bin_to_dec(val):
    return int(val, 2)

def new_string():
    global string_id
    
    string_id += 1
    return STRING_PATTERN%string_id

def suppress_output():
    devnull = open('/dev/null', 'w')
    oldstdout = os.dup(1)
    os.dup2(devnull.fileno(), 1)
    return (devnull, oldstdout)

def restore_output(saved_status):
    (devnull, old_stdout) = saved_status
    os.dup2(old_stdout, 1)
    devnull.close()

def sort_system_variables(variables, with_names=False):
    depthdic = {}
    maxdepth = 0
    for v in variables:
        varname = v.symbol_name()
        depth = len(varname.split("."))
        if depth > maxdepth:
            maxdepth = depth
        if depth not in depthdic:
            depthdic[depth] = []
        depthdic[depth].append((varname, v))

    ret = []
    depths = list(depthdic.keys())
    depths.sort()
    
    for i in depths:
        vars = depthdic[i]
        vars.sort()
        if with_names:
            ret += [v for v in depthdic[i]]
        else:
            ret += [v[1] for v in depthdic[i]]
    assert len(ret) == len(variables)
    return ret
        
def class_name(obj):
    return obj.__class__.__name__
    
class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'

def bold_text(text):
    return color.BOLD + text + color.END



# Copyright 2018 Cristian Mattarei
#
# Licensed under the modified BSD (3-clause BSD) License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from pysmt.shortcuts import And, Or, TRUE, FALSE, Not, EqualsOrIff, Implies, Iff, Symbol, BOOL, \
    BV, BVAdd, BVSub, get_type, BVULT, BVUGT, BVULE, BVUGE, Ite
from pysmt.typing import BOOL, BVType, ArrayType
from pysmt.fnode import FNode

from cosa.encoders.template import ClockBehavior
from cosa.representation import TS
from cosa.utils.logger import Logger
from cosa.encoders.formulae import StringParser
from cosa.utils.formula_mngm import BV2B

CLOCK_COUNTER = "__counter__"

class DeterministicClockBehavior(ClockBehavior):
    name = "DetClock"
    description = "Deterministic alternation of a defined period"
    interface = "clock-var, period"

    def get_default(self, params):
        return self.get_sts([params[0], 1])
    
    def get_sts(self, params):
        if len(params) != len(self.interface.split()):
            Logger.error("Invalid parameters for clock behavior \"%s\""%(self.name))
        clk = params[0]
        cyclestr = params[1]

        try:
            cycle = int(cyclestr)
        except:
            Logger.error("Clock cycle should be an integer number instead of \"%s\""%cyclestr)

        if (not type(clk) == FNode) or (not clk.is_symbol()):
            Logger.error("Clock symbol \"%s\" not found"%(str(clk)))

        init = []
        invar = []
        trans = []
        vars = set([])
        
        if clk.symbol_type().is_bv_type():
            pos_clk = EqualsOrIff(clk, BV(1 ,1))
            neg_clk = EqualsOrIff(clk, BV(0 ,1))
        else:
            pos_clk = clk
            neg_clk = Not(clk)
            
        if cycle < 1:
            Logger.error("Deterministic clock requires at least a cycle of size 1")
            
        if cycle == 1:
            init.append(neg_clk)
            trans.append(Iff(neg_clk, TS.to_next(pos_clk)))

        if cycle > 1:
            statesize = math.ceil(math.log(cycle)/math.log(2))
            counter = Symbol("%s%s"%(clk.symbol_name(), CLOCK_COUNTER), BVType(statesize))
            # 0 counts
            cycle -= 1

            # counter = 0 & clk = 0
            init.append(EqualsOrIff(counter, BV(0, statesize)))
            init.append(neg_clk)

            # counter <= cycle
            invar.append(BVULE(counter, BV(cycle, statesize)))

            # (counter < cycle) -> next(counter) = counter + 1
            trans.append(Implies(BVULT(counter, BV(cycle, statesize)), EqualsOrIff(TS.to_next(counter), BVAdd(counter, BV(1, statesize)))))
            
            # (counter >= cycle) -> next(counter) = 0
            trans.append(Implies(BVUGE(counter, BV(cycle, statesize)), EqualsOrIff(TS.to_next(counter), BV(0, statesize))))

            # (!clk) & (counter < cycle) -> next(!clk)
            trans.append(Implies(And(neg_clk, BVULT(counter, BV(cycle, statesize))), TS.to_next(neg_clk)))
            # (!clk) & (counter >= cycle) -> next(clk)
            trans.append(Implies(And(neg_clk, BVUGE(counter, BV(cycle, statesize))), TS.to_next(pos_clk)))

            # (clk) & (counter < cycle) -> next(clk)
            trans.append(Implies(And(pos_clk, BVULT(counter, BV(cycle, statesize))), TS.to_next(pos_clk)))
            # (clk) & (counter >= cycle) -> next(!clk)
            trans.append(Implies(And(pos_clk, BVUGE(counter, BV(cycle, statesize))), TS.to_next(neg_clk)))

            vars.add(counter)
            
        ts = TS("Clock Behavior")
        ts.vars, ts.init, ts.invar, ts.trans = vars, And(init), And(invar), And(trans)

        Logger.log("Adding clock behavior \"%s(%s)\""%(self.name, ", ".join([str(p) for p in params])), 1)
        
        return ts

class NondeterministicClockBehavior(ClockBehavior):
    name = "NondetClock"
    description = "Nondeterministic alternation of a maximum defined period"
    interface = "clock-var, period"

    def get_default(self, params):
        return self.get_sts([params[0], 3])
    
    def get_sts(self, params):
        if len(params) != len(self.interface.split()):
            Logger.error("Invalid parameters for clock behavior \"%s\""%(self.name))
        clk = params[0]
        cyclestr = params[1]

        try:
            cycle = int(cyclestr)
        except:
            Logger.error("Clock cycle should be an integer number instead of \"%s\""%cyclestr)

        if (not type(clk) == FNode) or (not clk.is_symbol()):
            Logger.error("Clock symbol \"%s\" not found"%(str(clk)))

        init = []
        invar = []
        trans = []
        vars = set([])
        
        if clk.symbol_type().is_bv_type():
            pos_clk = EqualsOrIff(clk, BV(1 ,1))
            neg_clk = EqualsOrIff(clk, BV(0 ,1))
        else:
            pos_clk = clk
            neg_clk = Not(clk)
            
        if cycle < 1:
            Logger.error("Deterministic clock requires at least a cycle of size 1")
            
        if cycle == 1:
            init.append(neg_clk)
            trans.append(Iff(neg_clk, TS.to_next(pos_clk)))

        if cycle > 1:
            statesize = math.ceil(math.log(cycle)/math.log(2))
            counter = Symbol("%s%s"%(clk.symbol_name(), CLOCK_COUNTER), BVType(statesize))
            # 0 counts
            cycle -= 1

            # counter = 0 & clk = 0
            init.append(EqualsOrIff(counter, BV(0, statesize)))
            init.append(neg_clk)

            # counter <= cycle
            invar.append(BVULE(counter, BV(cycle, statesize)))


            # (!clk) & (counter < cycle) -> ((next(!clk) & (next(counter) = counter + 1)) | (next(clk) & (next(counter) = 0)))
            trans.append(Implies(And(neg_clk, BVULE(counter, BV(cycle, statesize))), \
                                 Or(And(TS.to_next(neg_clk), EqualsOrIff(TS.to_next(counter), BVAdd(counter, BV(1, statesize)))), \
                                    And(TS.to_next(pos_clk), EqualsOrIff(TS.to_next(counter), BV(0, statesize))))))

            # (clk) & (counter < cycle) -> ((next(clk) & (next(counter) = counter + 1)) | (next(!clk) & (next(counter) = 0)))
            trans.append(Implies(And(pos_clk, BVULE(counter, BV(cycle, statesize))), \
                                 Or(And(TS.to_next(pos_clk), EqualsOrIff(TS.to_next(counter), BVAdd(counter, BV(1, statesize)))), \
                                    And(TS.to_next(neg_clk), EqualsOrIff(TS.to_next(counter), BV(0, statesize))))))

            # (counter >= cycle) -> ((next(clk) <-> !clk) & (next(counter) = 0))
            trans.append(Implies(BVUGE(counter, BV(cycle, statesize)), And(EqualsOrIff(TS.to_next(pos_clk), neg_clk), EqualsOrIff(TS.to_next(counter), BV(0, statesize)))))

            vars.add(counter)
            
        ts = TS("Clock Behavior")
        ts.vars, ts.init, ts.invar, ts.trans = vars, And(init), And(invar), And(trans)

        Logger.log("Adding clock behavior \"%s(%s)\""%(self.name, ", ".join([str(p) for p in params])), 1)
        
        return ts
    
class ConstantClockBehavior(ClockBehavior):
    name = "ConstClock"
    description = "Constant value"
    interface = "clock-var, value"

    def get_default(self, params):
        return self.get_sts(params)
    
    def get_sts(self, params):
        if len(params) != len(self.interface.split()):
            Logger.error("Invalid parameters for clock behavior \"%s\""%(self.name))
        clk = params[0]
        valuepar = params[1]

        if (not type(clk) == FNode) or (not clk.is_symbol()):
            Logger.error("Clock symbol \"%s\" not found"%(str(clk)))

        if (type(valuepar) == FNode) and (valuepar.is_bv_constant()):
            value = valuepar.constant_value()
        else:            
            try:
                value = int(valuepar)
            except:
                Logger.error("Clock value should be an integer number instead of \"%s\""%valuepar)

        init = []
        invar = []
        trans = []
        vars = set([])

        if clk.symbol_type().is_bv_type():
            pos_clk = EqualsOrIff(clk, BV(1, 1))
            neg_clk = EqualsOrIff(clk, BV(0, 1))
        else:
            pos_clk = clk
            neg_clk = Not(clk)

        if value == 1:
            invar.append(pos_clk)
        else:
            invar.append(neg_clk)
            
        ts = TS("Clock Behavior")
        ts.vars, ts.init, ts.invar, ts.trans = vars, And(init), And(invar), And(trans)

        Logger.log("Adding clock behavior \"%s(%s)\""%(self.name, ", ".join([str(p) for p in params])), 1)
        
        return ts
    
